/*  initctl.c

    Source file for  initctl  (init(8) control tool).

    Copyright (C) 2000  Richard Gooch

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Richard Gooch may be reached by email at  rgooch@atnf.csiro.au
    The postal address is:
      Richard Gooch, c/o ATNF, P. O. Box 76, Epping, N.S.W., 2121, Australia.
*/

/*
    This tool will send control messages to init(8). For example, it may
    request init(8) to start a service and will wait for that service to be
    available. If the service is already available, init(8) will not start it
    again.
    This tool may also be used to inspect the list of currently available
    services.


    Written by      Richard Gooch   28-FEB-2000

    Updated by      Richard Gooch   11-OCT-2000: Added provide support.

    Updated by      Richard Gooch   6-NOV-2000: Renamed to initctl.c
    
    Last Updated by Matthias Benkmann 9 May 2001

18 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
- added -u switch to indicate communication with a user space init.

3 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
- fixed FIFO creation to include user write permission. It only worked before
  because root can write to everything. That doesn't work for user space init.
  
7 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
- enter user space mode automatically when writing to /dev/initctl not allowed

9 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
- added -c(lear-service) 

10 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
- added -f(orce) switch

11 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
- added SIG_ROLLBACK

8 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
- added SIG_NO_INFO

16 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
- open initctl fifo O_NONBLOCK and then clear O_NONBLOCK afterwards. This
  makes sure that initctl will abort with an error instead of hanging if the
  other end of the fifo is not open (i.e. init is not responsive).

19 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
- added "--help" parameter, changed usage output.

23 Apr 2001 Matthias Benkmann <m.s.b@gmx.net>
- if USERSPACE_INIT env var is set then user space init is contacted

6 Sep 2001 Matthias Benkmann <m.s.b@gmx.net>
- if user space init is to be contacted, use $INIT_ROOT rather than $HOME

KNOWN BUGS that are too obscure to document in manpage:  
- if fork() or fopen() fails in init, display-services will block 
  for ever. Likewise if fopen() succeeds in init and fails in initctl, the
  init-spawned process will block for ever.
  A fix would be to open fifo non-blocking in initctl before sending
  COMMAND_DUMP list and to open non-blocking in init, too. initctl could use
  select() to wait for the connection and abort after a timeout.
- if COMMAND_CLEAR is used on an unavailable service with pending needers,
  these needers will block indefinitely. Not worth fixing. If
  someone uses COMMAND_CLEAR he should know what he's doing.
  [UPDATE 24 Feb 2001] Situation can not occur anymore as pending needers
  for unavailable services are not allowed anymore
*/
#include <unistd.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include "simpleinit.h"


static void signal_handler (int sig);
static void fix_userspace_paths(void);
static void output_usage(void);

static int caught_signal = 0;
static int userspace = 0;
static char initctl_name[PATH_MAX]="/dev/initctl";
static char pidfile_name[PATH_MAX]="/dev/init.pid";

int main (int argc, char **argv)
{
    int fd, nbytes;
    struct sigaction sa;
    sigset_t ss;
    char *ptr;
    long buffer[COMMAND_SIZE / sizeof (long) + 1];
    struct command_struct *command = (struct command_struct *) buffer;
    unsigned command_flags=0;
    pid_t init_pid=1;

    sigemptyset (&ss);
    sigaddset (&ss, SIG_PRESENT);
    sigaddset (&ss, SIG_NOT_PRESENT);
    sigaddset (&ss, SIG_FAILED);
    sigaddset (&ss, SIG_ROLLBACK);
    sigprocmask (SIG_BLOCK, &ss, NULL);
    sigemptyset (&sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = signal_handler;
    sigaction (SIG_PRESENT, &sa, NULL);
    sigaction (SIG_NOT_PRESENT, &sa, NULL);
    sigaction (SIG_FAILED, &sa, NULL);
    sigaction (SIG_ROLLBACK, &sa, NULL);
    command->pid = getpid ();
    command->ppid = getppid ();
    if ( ( ptr = strrchr (argv[0], '/') ) == NULL ) ptr = argv[0];
    else ++ptr;
    /*  First generate command number by looking at invocation name  */
    if (strcmp (ptr, "display-services") == 0)
	command->command = COMMAND_DUMP_LIST;
    else if (strcmp (ptr, "need") == 0) command->command = COMMAND_NEED;
    else if (strcmp (ptr, "provide") == 0) command->command = COMMAND_PROVIDE;
    else command->command = COMMAND_TEST;
    /*  Now check for switches  */
    while ( (argc > 1) && (argv[1][0] == '-') )
    {
	if (strcmp(argv[1],"-n") == 0)
	{
	    command->command = COMMAND_NEED;
	} else 
	if (strcmp(argv[1],"-r") == 0)
	{
	    command->command = COMMAND_ROLLBACK;
	} else
	if (strcmp(argv[1],"-d") == 0)
	{
	    command->command = COMMAND_DUMP_LIST;
	} else
	if (strcmp(argv[1],"-p") == 0)
	{
	    command->command = COMMAND_PROVIDE;
	} else
	if (strcmp(argv[1],"-c") == 0)
	{
	    command->command = COMMAND_CLEAR;
	} else
	if (strcmp(argv[1],"-f") == 0)
	{
	    command_flags |= FLAG_FORCE;
	} else
	if (strcmp(argv[1],"-u") == 0)
	{
	    userspace=1;
	} else
	if (strcmp(argv[1],"--help") == 0)
	{
	    output_usage();
	    exit(1);
	} else
	{
	    fprintf (stderr, "Illegal switch: \"%s\"\n", argv[1]);
	    exit (1);
	}
	--argc;
	++argv;
    }

    /* test for R_OK instead of W_OK so that it will work on read-only fs */
    if ((access(initctl_name,R_OK)!=0) || (getenv("USERSPACE_INIT")!=NULL)) userspace=1;
    if (userspace) fix_userspace_paths();
    
    switch (command->command)
    {
      case COMMAND_CLEAR:
      case COMMAND_TEST:
      case COMMAND_NEED:
      case COMMAND_PROVIDE:
	if (argc < 2) {output_usage(); exit(1);}
	/*  Fall through  */
      case COMMAND_ROLLBACK:
	if (argc > 1) strcpy (command->name, argv[1]);
	else command->name[0] = '\0';
	break;
      case COMMAND_DUMP_LIST:
	if (tmpnam (command->name) == NULL)
	{
	    fprintf (stderr, "Unable to create a unique filename\t%s\n",
		     ERRSTRING);
	    exit (1);
	}
	if (mkfifo (command->name, S_IRUSR|S_IWUSR) != 0)
	{
	    fprintf (stderr, "Unable to create FIFO: \"%s\"\t%s\n",
		     command->name, ERRSTRING);
	    exit (1);
	}
	break;
    }
    
    
    {
      long arg;
      /*The following open() MUST use O_WRONLY|O_NONBLOCK to notice if the
      other end of the fifo is not open (i.e. init is not responsive). Note
      especially that opening O_RDWR will *not* suit this purpose 
      (see fifo(4) )*/
      if ( ( fd = open (initctl_name, O_WRONLY|O_NONBLOCK, 0) ) < 0 )
      {
	  fprintf (stderr, "Error opening %s: %s\n", initctl_name, ERRSTRING);
	  exit (1);
      }
      /*Now clear the O_NONBLOCK flag*/
      arg=fcntl(fd,F_GETFL,0);
      arg=arg & ~O_NONBLOCK;
      fcntl(fd,F_SETFL,arg);
    }  
    if (userspace){
	  int fd;
	  if (((fd=open(pidfile_name,O_RDONLY)) < 0) ||
	      (read(fd,&init_pid,sizeof(init_pid)) < (ssize_t)sizeof(init_pid) )
	     ) {fprintf(stderr,"Error reading init.pid\n"); exit(1); }
	  close(fd);   
    }
    
    command->command |= command_flags;
    if (write (fd, buffer, COMMAND_SIZE) < COMMAND_SIZE)
    {
	fprintf (stderr, "Error writing %s: %s\n", initctl_name, ERRSTRING);
	exit (1);
    }
    close (fd);
    kill(init_pid,SIGCHLD); /* wake up init if sleeping, to process command */
    command->command &= CLEAR_FLAGS;
    
    if (command->command != COMMAND_DUMP_LIST)
    {
	sigemptyset (&ss);
	while (caught_signal == 0) sigsuspend (&ss);
	switch (command->command)
	{
	  case COMMAND_PROVIDE:
	    switch (caught_signal)
	    {
	      case SIG_PRESENT:
		return 1;
	      case SIG_NOT_PRESENT:
		return 0;
	      case SIG_NOT_CHILD:
		return 2;
	      case SIG_ROLLBACK:
	      default:
		return 3;
	    }
	    break;
	  default:
	    switch (caught_signal)
	    {
	      case SIG_PRESENT:
		return 0;
	      case SIG_NOT_PRESENT:
		return 2;
	      case SIG_FAILED:
		return 1;
	      case SIG_ROLLBACK:	
	    /*  case SIG_NO_INFO: same as SIG_ROLLBACK */
	      default:
		return 3;
	    }
	    break;
	}
	return 3;
    }
    /*  Read back the data and display it  */
    /*  The open call blocks until the other end is opened, too */
    if ( ( fd = open (command->name, O_RDONLY, 0) ) < 0 )
    {
	fprintf (stderr, "Error opening:\"%s\"\t%s\n",
		 command->name, ERRSTRING);
	exit (1);
    }
    unlink (command->name); /* fifo will keep working till fd is closed */
    fflush (stdout);
    while ( ( nbytes = read (fd, buffer, COMMAND_SIZE) ) > 0 )
	write (1, buffer, nbytes);
    close (fd);
    return (0);
}   /*  End Function main  */

static void signal_handler (int sig)
{
    caught_signal = sig;
}   /*  End Function signal_handler  */

static void fix_userspace_paths(void)
{
  int len;
  char* p;
  const char* homedir;
  char buffy[1024];

  homedir=getenv("INIT_ROOT");
  if (homedir==NULL)
  {
      fprintf ( stderr, "INIT_ROOT environment variable not set, using current directory!\n");
      if (getcwd(buffy,1024)==NULL)
      {
        fprintf ( stderr, "cannot determine current directory's name\n" );
        exit(1);
      }
      homedir=buffy;
  }
  len=strlen(homedir);
  
  for (p=strchr(initctl_name,0); (p+1)!=initctl_name; --p) p[len]=p[0];
  memcpy(initctl_name,homedir,len);
  for (p=strchr(pidfile_name,0); (p+1)!=pidfile_name; --p) p[len]=p[0];
  memcpy(pidfile_name,homedir,len);
}

static void output_usage(void)
{
  fprintf(stderr,"USAGE:\n");
  fprintf(stderr,"initctl [ -c | -d | -n | -p | -r ] [ -f ] [ -u ] service\n");
  fprintf(stderr,"need [ -f ] [ -r ] service\n");
  fprintf(stderr,"provide service\n");
  fprintf(stderr,"display-services\n");
  fprintf(stderr,"initctl service\n");
  fprintf(stderr,"initctl -c service\n");
}
