/* shutdown.c - shutdown a Linux system
 * Initially written by poe@daimi.aau.dk 
 * Currently maintained at ftp://ftp.daimi.aau.dk/pub/Software/Linux/
 */

/*
 * Modified by jrs@world.std.com to try to exec "umount -a" and if
 * that doesn't work, then umount filesystems ourselves in reverse
 * order.  The old-way was in forward order.  Also if the device
 * field of the mtab does not start with a "/" then give umount
 * the mount point instead.  This is needed for the nfs and proc
 * filesystems and yet is compatible with older systems.
 *
 * We also use the mntent library interface to read the mtab file
 * instead of trying to parse it directly and no longer give a
 * warning about not being able to umount the root.
 *
 * The reason "umount -a" should be tried first is because it may do
 * special processing for some filesystems (such as informing an
 * nfs server about nfs umounts) that we don't want to cope with here.
 */

/*
 * Various changes and additions to resemble SunOS 4 shutdown/reboot/halt(8)
 * more closely by Scott Telford (s.telford@ed.ac.uk) 93/05/18.
 * (I butchered Scotts patches somewhat. - poe)
 *
 * Changes by Richard Gooch <rgooch@atnf.csiro.au> (butchered by aeb)
 * introducing shutdown.conf.
 *
 * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
 * - added Native Language Support
 *
 * 2000-03-02 Richard Gooch <rgooch@atnf.csiro.au>
 * - pause forever if (pid == 1) and send SIGQUIT to pid = 1
 *
 * 2000-11-04 Richard Gooch <rgooch@atnf.csiro.au>
 * - continue reaping if (pid == 1)
 *
 * 2000-11-06 Richard Gooch <rgooch@atnf.csiro.au>
 * - shut down "finalprog" from /etc/inittab
 * - kill normal user (non-root and non-daemon) processes first with SIGTERM
 *
 * 2000-11-08 Richard Gooch <rgooch@atnf.csiro.au>
 * - rollback services
 * - do not unmount devfs (otherwise get harmless but annoying messages)
 * - created syncwait() for faster shutting down
 * - kill getty processes
 *
 * 2000-12-13 Matthias Benkmann <m.s.b@gmx.net>
 * - fixed memory allocation in kill_mortals  
 * - The line line[strlen (line) - 1] = '\0'; in kill_finalprog will
 *   cut off the last character if inittab is not terminated by a whitespace.
 *   Why is the line there anyway? fgets() already 0-terminates the string
 *   and isspace() reacts to nl, too so there's no reason to cut off nl.
 *   (the same thing appears in other places)
 * - What if finalprog doesn't terminate? There should be a timeout in
 *   stop_finalprog().
 * - wrote read_config_file() and added option "kill_delay" to specify how
 *   many seconds to sleep between SIGTERM and SIGKILL.
 * - added exec_with_params() so that halt_action can specify parameters
 * - added a few messages to tell the user what's going on (e.g. filesystems
 *   being unmounted with umount)
 *
 * 2000-12-14 Matthias Benkmann <m.s.b@gmx.net>
 * - added default_message option for shutdown.conf
 * - increased size of message buffer to match that of default_message
 *
 * 2000-12-14 Matthias Benkmann <m.s.b@gmx.net>
 * - added -f option to initctl call to ensure proper shutdown even if one
 *   or more stop scripts fail.
 *
 * 2001-2-16 Matthias Benkmann <m.s.b@gmx.net>
 * - added check if /dev/initctl can be written to (especially if it is opened 
 *   for reading by someone). This is a rough check
 *   for the presence of simpleinit (anybody know a better way?) to prevent
 *   rolling back services and sending of signals in case init is something
 *   else (e.g. /bin/sh which is *not* always bash and thus might choke on
 *   being sent SIGTSTOP, SIGTERM or SIGQUIT)
 *
 * 2001-05-12 Richard Gooch <rgooch@atnf.csiro.au>
 * - unblock all signals (sigmask from simpleinit(8) stopped sleep(3))
 * - close all files
 *
 * 2001-2-16 Matthias Benkmann <m.s.b@gmx.net>
 * - use /proc/mounts rather than /etc/mtab if possible. mtab is too easy to
 *   corrupt
 * - changed printf to my_puts() in unmount_disks_ourselves()
 */
 
 

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <utmp.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <termios.h>
#include <mntent.h>
#include <sys/mount.h>
#include <sys/wait.h>
#include <syslog.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include "linux_reboot.h"
#include "pathnames.h"
#include "xstrncpy.h"
#include "nls.h"

static void usage(void), int_handler(int), write_user(struct utmp *);
static void wall(void), write_wtmp(void), unmount_disks(void);
static void unmount_disks_ourselves(void);
static void swap_off(void), do_halt(char *);
static void kill_mortals (int sig);
static void stop_finalprog (void);
static void syncwait (int timeval);
static void read_config_file(void);
static char* fix_parameter(char* cline, const char* id);
static int exec_with_params(char* path);

char	*prog;		/* name of the program */
int	opt_reboot;	/* true if -r option or reboot command */
int	timeout;	/* number of seconds to shutdown */
int	opt_quiet;	/* true if no message is wanted */
int	opt_fast;	/* true if fast boot */
char	message[256];	/* reason for shutdown if any... */
char 	default_message[256]; /*default message read from _PATH_SHUTDOWN_CONF*/
int	opt_single = 0; /* true is we want to boot singleuser */
char	*whom;		/* who is shutting the system down */
int	opt_msgset = 0; /* message set on command line */
			/* change 1 to 0 if no file is to be used by default */
int	defmsgset = 0;  /* default message set in _PATH_SHUTDOWN_CONF */			
int	opt_use_config_file = 1;	/* read _PATH_SHUTDOWN_CONF */
char	halt_action[256];		/* to find out what to do upon halt */
int 	kill_delay = 2;			/* delay between SIGTERM and SIGKILL*/

char	initctl_name[]="/dev/initctl";
int 	init_is_simpleinit= 1;  /*will be 0 if init is /bin/sh or sth else*/
char	ghost[]="ghost";

/* #define DEBUGGING */

#define WR(s) write(fd, s, strlen(s))
#define WRCRLF	write(fd, "\r\n", 2)
#define ERRSTRING strerror(errno)
#define MAXPARAMS 128  /*max. number of params you can pass to halt_action prog*/

void
usage(void)
{
	fprintf(stderr,
		_("Usage: shutdown [-h|-r] [-fqs] [now|hh:ss|+mins]\n"));
	exit(1);
}

static void
my_puts(const char *s)
{
	/* Use a fresh stdout after forking */
	freopen(_PATH_CONSOLE, "w", stdout);
	puts(s);
	fflush(stdout);
}

void 
int_handler(int sig)
{
	unlink(_PATH_NOLOGIN);
	signal(SIGINT, SIG_DFL);
	my_puts(_("Shutdown process aborted"));
	exit(1);
	sig=sig;
}

int
main(int argc, char *argv[])
{
	int c, i, fd;	
	char *ptr;

	i = getdtablesize ();
	for (fd = 3; fd < i; fd++) close (fd);
	if (getpid () == 1)
	{
	    for (fd = 0; fd < 3; fd++) close (fd);
	    while (1) wait (NULL);  /*  Grim reaper never stops  */
	}
	sigsetmask (0); /*  simpleinit(8) blocks all signals: undo for ALRM  */
	for (i = 1; i < NSIG; i++) signal (i, SIG_DFL); /* unnecessary */
        setlocale(LC_ALL, "");
        bindtextdomain(PACKAGE, LOCALEDIR);
        textdomain(PACKAGE);

#ifndef DEBUGGING
	if(setreuid (0, 0)) {
		fprintf(stderr, _("%s: Only root can shut a system down.\n"),
			argv[0]);
		exit(1);
	}
#endif

	if(*argv[0] == '-') argv[0]++;	/* allow shutdown as login shell */
	prog = argv[0];
	if((ptr = strrchr(argv[0], '/'))) prog = ++ptr;

	/* All names (halt, reboot, fasthalt, fastboot, shutdown)
	   refer to the same program with the same options,
	   only the defaults differ. */
	if(!strcmp("halt", prog)) {
		opt_reboot = 0;
		opt_quiet = 1;
		opt_fast = 0;
		timeout = 0;
	} else if(!strcmp("fasthalt", prog)) {
		opt_reboot = 0;
		opt_quiet = 1;
		opt_fast = 1;
		timeout = 0;
	} else if(!strcmp("reboot", prog)) {
		opt_reboot = 1;
		opt_quiet = 1;
		opt_fast = 0;
		timeout = 0;
	} else if(!strcmp("fastboot", prog)) {
		opt_reboot = 1;
		opt_quiet = 1;
		opt_fast = 1;
		timeout = 0;
	} else {
		/* defaults */
		opt_reboot = 0;
		opt_quiet = 0;
		opt_fast = 0;
		timeout = 2*60;
	}
		
	c = 0;
	while(++c < argc) {
		if(argv[c][0] == '-') {
			for(i = 1; argv[c][i]; i++) {
				switch(argv[c][i]) {
				case 'C':
					opt_use_config_file = 1;
					break;
				case 'h': 
					opt_reboot = 0;
					break;
				case 'r':
					opt_reboot = 1;
					break;
				case 'f':
					opt_fast = 1;
					break;
				case 'q':
					opt_quiet = 1;
					break;
				case 's':
					opt_single = 1;
					break;
				    
				default:
					usage();
				}
			}
		} else if(!strcmp("now", argv[c])) {
			timeout = 0;
		} else if(argv[c][0] == '+') {
			timeout = 60 * atoi(&argv[c][1]);
		} else if (isdigit(argv[c][0])) {
			char *colon;
			int hour = 0;
			int minute = 0;
			time_t tics;
			struct tm *tt;
			int now, then;
				
			if((colon = strchr(argv[c], ':'))) {
				*colon = '\0';
				hour = atoi(argv[c]);
				minute = atoi(++colon);
			} else usage();
				
			(void) time(&tics);
			tt = localtime(&tics);
				
			now = 3600 * tt->tm_hour + 60 * tt->tm_min;
			then = 3600 * hour + 60 * minute;
			timeout = then - now;
			if(timeout < 0) {
				fprintf(stderr, _("That must be tomorrow, "
					          "can't you wait till then?\n"));
				exit(1);
			}
		} else {
			xstrncpy(message, argv[c], sizeof(message));
			opt_msgset = 1;
		}
	}

        read_config_file();
	
	if(!opt_quiet && !opt_msgset) {
		/* now ask for message, gets() is insecure */
		int cnt = sizeof(message)-1;
		char *ptr;
		
		printf("Why? "); fflush(stdout);
		
		ptr = message;
		while(--cnt >= 0 && (*ptr = getchar()) && *ptr != '\n') { 
			ptr++;
		}
		*ptr = '\0';
	} else if (!opt_msgset) {
	  	if (!defmsgset) 
	  	  strcpy(message, _("for maintenance; bounce, bounce"));
	  	else 
	  	  strcpy(message, default_message);
	}

#ifdef DEBUGGING
	printf(_("timeout = %d, quiet = %d, reboot = %d\n"),
		timeout, opt_quiet, opt_reboot);
#endif
	
	/* so much for option-processing, now begin termination... */
	if(!(whom = getlogin()) || !*whom) whom = ghost;
	if(strlen(whom) > 40) whom[40] = 0; /* see write_user() */

	setpriority(PRIO_PROCESS, 0, PRIO_MIN);
	signal(SIGINT,  int_handler);
	signal(SIGHUP,  int_handler);
	signal(SIGQUIT, int_handler);
	signal(SIGTERM, int_handler);

	chdir("/");

	if(timeout > 5*60) {
		sleep(timeout - 5*60);
		timeout = 5*60;
	}

	
	if((fd = open(_PATH_NOLOGIN, O_WRONLY|O_CREAT, 0644)) >= 0) {
		/* keep xgettext happy and leave \r\n outside strings */
		WRCRLF;
		WR(_("The system is being shut down within 5 minutes"));
		WRCRLF;
		write(fd, message, strlen(message));
		WRCRLF;
		WR(_("Login is therefore prohibited."));
		WRCRLF;
		close(fd);
	}
	
	signal(SIGPIPE, SIG_IGN);

	if(timeout > 0) {
		wall();
		sleep(timeout);
	}

	timeout = 0;
	wall();
	sleep(3);

	/* now there's no turning back... */
	signal(SIGINT,  SIG_IGN);

	/* do syslog message... */
	openlog(prog, LOG_CONS, LOG_AUTH);
	syslog(LOG_NOTICE, _("%s by %s: %s"), 
	       opt_reboot ? _("rebooted") : _("halted"), whom, message);
	closelog();

	if(opt_fast)
		if((fd = open("/fastboot", O_WRONLY|O_CREAT, 0644)) >= 0)
			close(fd);

	/*Check if communication via /dev/initctl is possible. If not, assume
	that init is not simpleinit (could be /bin/sh for instance).*/
	{
	  /*must be O_WRONLY|O_NONBLOCK to catch situation that fifo is
	  not opened on the other side (see fifo(4) ).*/
	  int fd=open(initctl_name,O_WRONLY|O_NONBLOCK,0);
	  if (fd<0) init_is_simpleinit=0; else close(fd);
	}

	/* tell init not to spawn more gettys */
	if (init_is_simpleinit) kill(1, SIGTSTP);
	
	write_wtmp();
	if(opt_single)
		if((fd = open(_PATH_SINGLE, O_CREAT|O_WRONLY, 0644)) >= 0)
			close(fd);
		
	sync();

	signal(SIGTERM, SIG_IGN);
	if(fork() > 0) sleep(1000); /* the parent will die soon... */
	setpgrp();		/* so the shell wont kill us in the fall */

#ifndef DEBUGGING
	my_puts ("");
	my_puts ("");                 /*  Get past the login prompt         */
	/* a gentle kill of all other processes except those with uid>=100 */
	/* the message stating this is in kill_mortals() because it is */
	/* only printed if some processes are actually killed.*/
	kill_mortals (SIGTERM);
	for (fd = 0; fd < 3; fd++) close (fd);
	stop_finalprog (); /*should this be done if init_is_simpleinit==0 ??*/
	sleep (1);                    /*  Time for saves to start           */
	if (init_is_simpleinit) 
		kill (1, SIGTERM);    /*  Tell init to kill gettys          */
	usleep (100000);              /*  Wait for gettys to die            */
	if (init_is_simpleinit) 
		system ("/sbin/initctl -r -f"); /*  Roll back services      */
	syncwait (1);
	
	my_puts ("Sending TERM signal to all remaining processes...");
	kill(-1, SIGTERM);
	sleep(kill_delay);	/* default 2, change via shutdown.conf */

	/* now use brute force... */
	my_puts ("Sending KILL signal to all remaining processes...");
	kill(-1, SIGKILL);

	/* turn off accounting */
	acct(NULL);
#endif
	/* RedHat and SuSE like to remove /etc/nologin.
	   Perhaps the usual sequence is
	      touch nologin; shutdown -h; fiddle with hardware;
	      boot; fiddle with software; rm nologin
	   and removing it here will be counterproductive.
	   Let us see whether people complain. */
	unlink(_PATH_NOLOGIN);

	sleep (1);	/* Time for processes to die and close files */
	syncwait (2);

	if (init_is_simpleinit)
	{
	  /*  Tell init(8) to exec so that the old inode may be freed cleanly if
	      required. Need to sleep before remounting root read-only  */
	  my_puts ("Telling init to exec shutdown...");
	  kill (1, SIGQUIT);
	}  

	/* remove swap files and partitions using swapoff */
	my_puts ("Removing swap files and partitions via swapoff...");
	swap_off();

	/* unmount disks... */
	my_puts ("Unmounting filesystems via umount...");
	unmount_disks();
	syncwait (1);

	if(opt_reboot) {
		my_puts(_("Please stand by while rebooting the system...\n"));
		sleep(1);
		my_reboot(LINUX_REBOOT_CMD_RESTART); /* RB_AUTOBOOT */
		my_puts(_("\nWhy am I still alive after reboot?"));
	} else {
		my_puts(_("\nNow you can turn off the power..."));

		/* allow C-A-D now, faith@cs.unc.edu, re-fixed 8-Jul-96 */
		my_reboot(LINUX_REBOOT_CMD_CAD_ON); /* RB_ENABLE_CAD */
		sleep (1);  /*  Wait for devices to finish writing to media  */
		do_halt(halt_action);
	}
	/* NOTREACHED */
	exit(0); /* to quiet gcc */
}

/*** end of main() ***/

/*takes a string like
/bin/sh -c "echo Goodbye World"
and calls execv("/bin/sh",{"sh","-c","echo Hallo World"})
The backslash \ is the escape character inside "...", whenever a 
\ is encountered it will be discarded and the following
char will be left alone. " inside "..." can be written \" and \ inside
"..." can be written \\. The backslash can *not* be used to embed NUL
characters. A backslash preceding a NUL character will be discarded and
the string terminated.
The function will not fail if the closing " are missing. The string is
properly terminated at the 0 character.
RETURN VALUE:
that of execv()
ATTENTION!! This function destroys the contents of its path argument!*/
int exec_with_params(char* path)
{ 			/*written by MSB*/
  char* argv[MAXPARAMS];
  int argc;
  
  for (argv[0]=path; (*argv[0]!=0) && !isspace(*argv[0]); ++argv[0]);
  while ((argv[0]!=path) && (*argv[0]!='/')) --argv[0];
  if (*argv[0]=='/') ++argv[0];
  
  argc=1;
  while((argc<MAXPARAMS) && (argv[argc-1]!=NULL))
  {
    /*find beginning of argv[argc] and at the same time fix argv[argc-1] 
    if it contains "..."*/
    argv[argc]=argv[argc-1];
    
    if (*argv[argc]=='"')
    {
      int move=1;
      ++argv[argc];
      for(;;)
      {
        if ((*argv[argc]==0) || (*argv[argc]=='"')) {
          argv[argc][-move]=0; 
          break;
        } else
        if (*argv[argc]=='\\')
        {
          ++move;
          ++argv[argc];
          if (*argv[argc]==0) continue; /*no embedded NULs allowed*/
        }
        argv[argc][-move]=*argv[argc];
        ++argv[argc];
      }
    } else while ((*argv[argc]!=0) && !isspace(*argv[argc])) ++argv[argc];
    
    /*terminate argv[argc-1] if necessary*/
    if (*argv[argc]!=0) {*argv[argc]=0; ++argv[argc];}

    /*fixing argv[argc-1] is finished, now look for start of argv[argc]*/
    while (isspace(*argv[argc])) ++argv[argc];
    if (*argv[argc]==0) argv[argc]=NULL; 
    ++argc;
  }  
  return execv(path,argv); 
}

void
do_halt(char *action) {
	if (strcasecmp (action, "power_off") == 0) {
		printf(_("Calling kernel power-off facility...\n"));
		fflush(stdout);
		my_reboot(LINUX_REBOOT_CMD_POWER_OFF);
		printf(_("Error powering off\t%s\n"), ERRSTRING);
		fflush(stdout);
		sleep (2);
	} else

	/* This should be improved; e.g. Mike Jagdis wants "/sbin/mdstop -a" */
	/* No problemo - MSB */
	/* Maybe we should also fork and wait */
	if (action[0] == '/') {
		printf(_("Executing the program \"%s\" ...\n"), action);
		fflush(stdout);
 		exec_with_params(action);
		printf(_("Error executing\t%s\n"), ERRSTRING);
		fflush(stdout);
		sleep (2);
	}

	my_reboot(LINUX_REBOOT_CMD_HALT); /* RB_HALT_SYSTEM */
}

void
write_user(struct utmp *ut)
{
	int fd;
	int minutes, hours;
	char term[40] = {'/','d','e','v','/',0};
	char msg[100];

	minutes = timeout / 60;
	hours = minutes / 60;
	minutes %= 60;
	
	(void) strncat(term, ut->ut_line, sizeof(ut->ut_line));

	/* try not to get stuck on a mangled ut_line entry... */
	if((fd = open(term, O_WRONLY|O_NONBLOCK)) < 0)
	        return;

	msg[0] = '\007';	/* gettext crashes on \a */
	sprintf(msg+1, _("URGENT: broadcast message from %s:"), whom);
	WRCRLF;
	WR(msg);
	WRCRLF;

	if (hours > 1)
		sprintf(msg, _("System going down in %d hours %d minutes"),
			hours, minutes);
	else if (hours == 1)
		sprintf(msg, _("System going down in 1 hour %d minutes"),
			minutes);
	else if (minutes > 1)
		sprintf(msg, _("System going down in %d minutes\n"),
			minutes);
	else if (minutes == 1)
		sprintf(msg, _("System going down in 1 minute\n"));
	else
		sprintf(msg, _("System going down IMMEDIATELY!\n"));
	
	WR(msg);
	WRCRLF;

	sprintf(msg, _("\t... %s ...\n"), message);
	WR(msg);
	WRCRLF;

	close(fd);
}

void
wall(void)
{
	/* write to all users, that the system is going down. */
	struct utmp *ut;
		
	utmpname(_PATH_UTMP);
	setutent();
	
	while((ut = getutent())) {
		if(ut->ut_type == USER_PROCESS)
			write_user(ut);
	}
	endutent();
}

void
write_wtmp(void)
{
	/* write in wtmp that we are dying */
	int fd;
	struct utmp ut;
	
	memset((char *)&ut, 0, sizeof(ut));
	strcpy(ut.ut_line, "~");
	memcpy(ut.ut_name, "shutdown", sizeof(ut.ut_name));

	time(&ut.ut_time);
	ut.ut_type = BOOT_TIME;
	
	if((fd = open(_PATH_WTMP, O_WRONLY|O_APPEND, 0644)) >= 0) {
		write(fd, (char *)&ut, sizeof(ut));
		close(fd);
	}
}

void
swap_off(void)
{
	/* swapoff esp. swap FILES so the underlying partition can be
	   unmounted. It you don't have swapoff(1) or use mount to
	   add swapspace, this may not be necessary, but I guess it
	   won't hurt */

	int pid;
	int result;
	int status;

	sync();
	if ((pid = fork()) < 0) {
		my_puts(_("Cannot fork for swapoff. Shrug!"));
		return;
	}
	if (!pid) {
		execl("/sbin/swapoff", SWAPOFF_ARGS, NULL);
		execl("/etc/swapoff", SWAPOFF_ARGS, NULL);
		execl("/bin/swapoff", SWAPOFF_ARGS, NULL);
		execlp("swapoff", SWAPOFF_ARGS, NULL);
		my_puts(_("Cannot exec swapoff, "
			  "hoping umount will do the trick."));
		exit(0);
	}
	while (((result = wait(&status)) != -1) && (result != pid))
		;
}

void
unmount_disks(void)
{
	/* better to use umount directly because it may be smarter than us */

	int pid;
	int result;
	int status;

	sync();
	if ((pid = fork()) < 0) {
		my_puts(_("Cannot fork for umount, trying manually."));
		unmount_disks_ourselves();
		return;
	}
	if (!pid) {
		execl(_PATH_UMOUNT, UMOUNT_ARGS, NULL);

		/* need my_printf instead of my_puts here */
		freopen(_PATH_CONSOLE, "w", stdout);
		printf(_("Cannot exec %s, trying umount.\n"), _PATH_UMOUNT);
		fflush(stdout);

		execlp("umount", UMOUNT_ARGS, NULL);
		my_puts(_("Cannot exec umount, giving up on umount."));
		exit(0);
	}
	while (((result = wait(&status)) != -1) && (result != pid))
		;
	my_puts(_("Unmounting remaining filesystems manually..."));
	unmount_disks_ourselves();
}

void
unmount_disks_ourselves(void)
{
	/* unmount all disks */

	FILE *mtab;
	struct mntent *mnt;
	char *mntlist[128];
	int i;
	int n;
	char *filesys;
	
	sync();
	if (!(mtab = setmntent(_PATH_PROC_MOUNTS, "r"))) {
	        char buffy[256];
	        sprintf(buffy,"shutdown: Cannot open " _PATH_PROC_MOUNTS ": %s",ERRSTRING);
	        my_puts(buffy);
	        if (!(mtab = setmntent(_PATH_MTAB, "r"))) {
	        	sprintf(buffy,"shutdown: Cannot open " _PATH_MTAB ": %s",ERRSTRING);
	        	my_puts(buffy);
			return;
		}	
	}
	n = 0;
	while (n < 100 && (mnt = getmntent(mtab))) {
		if (strcmp (mnt->mnt_type, "devfs") == 0) continue;
		mntlist[n++] = strdup(mnt->mnt_dir);
	}
	endmntent(mtab);

	/* we are careful to do this in reverse order of the mtab file */

	for (i = n - 1; i >= 0; i--) {
		filesys = mntlist[i];
#ifdef DEBUGGING
		printf("umount %s\n", filesys);
#else
		if (umount(mntlist[i]) < 0)
		{
			char buffy[1024];
			sprintf(buffy,_("shutdown: Couldn't umount %s: %s"),
			       filesys, ERRSTRING);
			my_puts(buffy);
		} 
		free(mntlist[i]); 
#endif
	}
}

static void kill_mortals (int sig)
{
    int npids = 0;
    int index = 0;
    int pid;
    struct stat statbuf;
    DIR *dp;
    struct dirent *de;
    pid_t *pids = NULL;
    pid_t* pids2;
    char path[256];

    if ( ( dp = opendir ("/proc") ) == NULL ) return;
    while ( ( de = readdir (dp) ) != NULL )
    {
	if ( !isdigit (de->d_name[0]) ) continue;
	pid = atoi (de->d_name);
	sprintf (path, "/proc/%d", pid);
	if (stat (path, &statbuf) != 0) continue;
	if (statbuf.st_uid < 100) continue;
	if (index >= npids)
	{
	    pids2 = realloc (pids, (npids+4096)*sizeof(pid_t));
	    if (pids2 == NULL) break;
	    pids=pids2;
	    npids += 4096;
	}
	pids[index++] = pid;
    }
    closedir (dp);
    if (pids == NULL) return;
    my_puts ("Sending TERM signal to processes with UID>=100...");
    for (--index; index >= 0; --index) kill (pids[index], sig);
    free (pids);
}   /*  End Function kill_mortals  */

static void stop_finalprog (void)
{
    char *p1, *p2;
    FILE *fp;
    char line[256];

    if ( ( fp = fopen (_PATH_INITTAB, "r") ) == NULL ) return;
    while (fgets (line, 256, fp) != NULL)
    {
	pid_t pid;

	line[strlen (line) - 1] = '\0';
	p1 = line;
	while ( isspace (*p1) ) ++p1;
	if (strncmp (p1, "finalprog", 9) != 0) continue;
	if ( ( p1 = strchr (p1 + 9, '=') ) == NULL ) continue;
	for (++p1; isspace (*p1); ++p1);
	if (*p1 == '\0') continue;
	for (p2 = p1; !isspace (*p2); ++p2);
	*p2 = '\0';
	switch ( pid = fork () )
	{
	  case 0:   /*  Child   */
	    execl (p1, p1, "stop", NULL);
	    break;
	  case -1:  /*  Error   */
	    break;
	  default:  /*  Parent  */
	    waitpid (pid, NULL, 0);
	    break;
	}
	fclose (fp);
	return;
    }
    fclose (fp);
}   /*  End Function stop_finalprog  */

static void syncwait (int timeval)
{
    static int do_wait = 0;
    static int first_time = 1;

    sync ();
    /*  Kernel version 1.3.20 and after are supposed to wait automatically  */
    if (first_time)
    {
	struct utsname uts;

	first_time = 0;
	uname (&uts);
	if (uts.release[0] < '2') do_wait = 1;
    }
    if (do_wait) sleep (timeval);
}   /*  End Function syncwait  */

/*tests if cline starts with an arbitrary number of whitespace, followed by id 
(case insensitive!!), followed by an arbitrary
number of = signs or whitespace. If so, then trailing whitespace is removed.
RETURN value:
 NULL if cline doesn't match id,
 pointer to processed string (i.e. component after =) otherwise
ATTENTION! The processed string may be empty (i.e. =="\0") if nothing but
ws or = signs followed id.
*/
static char* fix_parameter(char* cline, const char* id)
{            /*written by MSB*/
  char* str;
  int len=strlen(id);

  while (isspace(*cline)) ++cline;
  if (strncasecmp (cline,id,len)!=0) return NULL;
  if ((cline[len]!='=') && (!isspace(cline[len]))) return NULL; 
 
  for (cline+=len; isspace(*cline) || (*cline=='='); ++cline);
  if (*cline=='\0') return cline;
  
  /*cline is not empty ==> cut off trailing whitespace*/
  for (str=cline; *str!=0; ++str);
  for (--str; ((str!=cline) && isspace(*str)); --str);
  if (!isspace(*str)) ++str;
  *str=0;
  return cline;
}

/* No doubt we shall want to extend this some day
and register a series of commands to be executed
at various points during the shutdown sequence,
and to define the number of milliseconds to sleep, etc. */
static void read_config_file(void)
{   
  char line[256];
  FILE *fp;
  halt_action[0] = '\0';
  
  if (!opt_use_config_file) return; 
  
  /*  Read and parse the config file */
  if ((fp = fopen (_PATH_SHUTDOWN_CONF, "r")) == NULL) return;

  while (!feof(fp))
  {
    char* str;
    if (fgets(line,256,fp)==NULL) break;
    
    for (str=line; isspace(*str); ++str);
    if ((*str=='#') || (*str==0)) continue;
    
    if ((str=fix_parameter(line,"HALT_ACTION"))!=NULL) strcpy(halt_action, str);
    else 
    if ((str=fix_parameter(line,"KILL_DELAY"))!=NULL) 
    {
      char* endptr;
      long int l=strtol(str,&endptr,0);
      /*pedantic checking of input value because a wrong one can cause*/
      /*data loss! Maybe the value 0 should not even be allowed here but*/
      /*I guess this is root's responsibility so just make sure that*/
      /*invalid numbers are rejected.*/
      if ((*str!=0) && (*endptr==0) && (l>=0) && (l<=INT_MAX))
      							kill_delay=atoi(str);
    }
    else
    if ((str=fix_parameter(line,"DEFAULT_MESSAGE"))!=NULL) 
                           { strcpy(default_message, str); defmsgset=1; }
    /*ATTENTION! Be careful when adding new options. Don't forget that*/
    /*fix_parameter does *not* return NULL if the string is empty but id*/
    /*id is matched*/
  }
  fclose (fp);
}
