/* simpleinit.c - poe@daimi.aau.dk */
/* Version 2.0.1-msb */

/* 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
 * - added Native Language Support
 */

/* 12 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * -added console opening to need command (was already present for rollback)
 *  so that need executed manually will not swallow messages.
 * -Fixed bug in get_path (incorrect string concatenation) that prevented 
 *  init_path from working.
 * -added sigint_critical variable to make sure init's behaviour is not 
 *  changed when SIGINT is received after rc scripts have been run. This is
 *  a safety measure because in future someone might carelessly write code that
 *  uses caught_sigint which would make init behave differently after a SIGINT,
 *  even though a sigint is a normal event that does not always lead to shutting
 *  down the computer (the shutdown program might be replaced!).
 * -added quietrollback option to suppress "stopped service:" messages 
 * -fixed bug in read_inittab()  ("finalprog" has 9 letters, not 8!)
 *
 * 15 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * - added while(1) before sigsuspend() in do_rc_tty() just in case sth goes 
 *   wrong
 *
 * 16 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * - changed installation of SIGINT handler to use sigaction() instead of
 *   obsolete signal() (why was SIGINT still installed via signal() unlike
 *   the other handlers?), same for SIGHUP
 * - changed read_inittab() to use fix_parameter(). Note that this makes
 *   the parameters case-insensitive and = becomes optional.
 * - added error message for non-empty inittab lines that contain neither a 
 *   recognized command, nor a comment, nor a ":".
 * - Shouldn't "if (initctl_fd < 0) return waitpid (pid, status, 0);"
 *   have special treatment for (retval==-1) && (errno==EINTR) ?
 * - changed sigsetjmp call to save mask. Why else would I use that func?
 * 
 * 17 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * - added ability to run in user space; extremely useful for testing your
 *   bootscript setup, especially in chroot.
 * - added "ctrlaltdel=" option to inittab, to specify which program should
 *   be run when SIGINT is received by init. Note that neither the reboot
 *   program nor the syscall are used as fallback if exec'ing this prog fails
 *   because a user specifying a ctrlaltdel program probably does it because
 *   he does *not* want a reboot. Additionally this allows using the entry
 *   "ctrlaltdel=/sbin/reboot" to prevent the hard reboot fallback.
 *
 * 18 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * - added protection against nervous Ctrl-Alt-Del pressers 
 *
 * 26 Dec 2000 Matthias Benkmann <m.s.b@gmx.net>
 * - added code to clean up /var/run/utmp and to create INIT_PROCESS entries
 *   recycling utmp lines with the same ut_id. This fixes problems with 
 *   programs that use getutid() (or functions like pututline() that rely on
 *   getutid()) after setutent(), like shadowpwd's login for instance.
 *   It also prevents /var/run/utmp from growing unnecessarily.
 *
 * 1 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - introduced did_rc_tty variable to prevent running do_rc_tty() 2 times
 *   (1st in do_single and then regular).
 * - changed sigquit_handler() to execute _PATH_SHUTDOWN instead of
 *   _PATH_REBOOT, because the latter might have been replaced by a special
 *   script. It's generally unadvisable to rely on /sbin/reboot to be 
 *   a link to /sbin/shutdown.
 *
 * 9 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - added code to clean up next_attempting_provider entries by setting them 
 *   NULL once they're no longer used. Not strictly necessary but someone might
 *   add code in the future that expects unused entries to be NULL. Better safe
 *   than sorry.
 * - added fancy (IMHO) ASCII drawing to visualize pointer orgy used in
 *   service accounting.
 * - fixed memory leak: "free (victim)" would free a script from the 
 *   available_list and leave its services hanging.
 * - added COMMAND_CLEAR which removes the accounting entry for a service
 *   without stopping it (works for unavailable services).
 *
 * 10 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - mask signals before forking to make sure children have enough time to
 *   reset signal handlers. There wasn't really a problem but I like avoiding 
 *   the mere possibility of a race condition out of principle.
 * - replaced the weird longjmp hack to fix the shutdown/init 
 *   SIGQUIT/SIGCHLD race condition.
 *
 * 11 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - changed rollback handling so that init doesn't block anymore while
 *   stop script is being executed. This fixes the problem that any
 *   stop script using initctl would freeze init for good. A need -r 
 *   used while another rollback is in progress will be ignored if it specifies
 *   a service within the rollback range and will extend the rollback range
 *   otherwise, display-services is unaffected and can be used during rollback
 *   without restrictions.
 *
 * 12 Jan 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - implemented -f option for rollback (forced rollback)
 * - implemented -f watchdog that kills stop scripts that don't terminate
 *   within 5 minutes.
 *
 * 8 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - made COMMAND_TEST more verbose with its exit codes
 *
 * 21 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - fixed 2 bugs that caused the unavailable_list to become a loop, resulting
 *   in an endless loop in force_progress(). 
 * - fixed deadlock if a script calls provide multiple times for one service
 *
 * 23 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - fixed deadlock if script needs itself, need will return 0 in this case.
 *   This is useful when service names are hidden behind shell vars and 
 *   scripts are being merged, e.g. klogd and syslogd are merged into 
 *   logging. You don't have to remove "need $klogd" even though $klogd 
 *   resolves to logging now.
 *
 * 24 Feb 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - implemented full deadlock detection. Whenever a deadlock occurs, no
 *   no matter how obscure the combination of provide and need that caused it,
 *   init will output the names of the offending scripts. Detection only works
 *   for children of init. This change is somewhat incompatible as it causes
 *   some needs to fail early that would previously wait for signal_needers()
 *   in force_progress(). However, this only affects boot script setups
 *   that were fragile in the first place because they either used
 *   secondary services with the same name as primary services 
 *   (race condition, see BUGS in initctl(8))
 *   or use need with the name of a secondary service (implicit dependency).
 * - fixed bug that handle_nonworking() and process_pidstat() would kill
 *   the actual provider SCRIPT, instead of the initctl instance waiting
 *   desperately for its signal.
 * - fixed memory leak
 * - fixed problem that prevented rollback if bootprog is a directory that is
 *   not in init_path
 *
 * 28 Mar 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - fixed bug in mywait() that prevented proper zombie removal
 * - fixed bug in mywait() that prevented SIGQUIT from being registered
 *
 * 23 Apr 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - added INIT_D env var
 * - added USERSPACE_INIT env var
 *
 * 9 May 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - sync'ing with util-linux-2.11b: 
 *      renamed mywaitpid->mywait
 *      did not add ignore_dangling_symlink code (I see no reason)
 *      changed strncpy -> xstrncpy
 *      did not introduce new mywait() code (dirty hack; I'm not convinced
 *           it really fixes the race; my code is a lot cleaner)
 *      added find_script_byname() fix (although not necessary because 
 *           I did not change run_command)
 *      did not introduce incompatible change in run_command that disables
 *           retrying of failed services (too harsh; retry ensures best
 *           effort; if a script really does not want a retry it can use
 *           initctl to test for it first)
 *
 * 10 May 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - break while(getutent()) loops if pututline fails because it seems
 *   getutent() is stuck at the same entry if pututline fails which leads
 *   to an endless loop if rc scripts don't mount / read-write.
 *
 * 6 Sep 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - changed user space init to interpret paths relative to the directory
 *   where it was started rather than $HOME, unless INIT_ROOT is set
 * - export INIT_ROOT to point to user space init's root dir.
 *
 * 7 Sep 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - removed "single" special case. People should create a bootscript called
 *   "single" instead.
 * - removed /bin/sh fallback for security reasons.
 * - removed parallel execution capability (I prefer deterministic behaviour
 *   and don't want to maintain code I don't test)
 *   This breaks compatibility with R. Gooch's boot scheme.
 * - added parameter handling for bootprog
 *
 * 9 Sep 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - added ability to use directory names as services which are run via 
 *   bootprog.
 *
 * 12 Sep 2001 Matthias Benkmann <m.s.b@gmx.net>
 * - replaced wtmp handling code with a single call to updwtmp(). If this
 *   function is not available (libc is older than glibc2), wtmp handling
 *   is disabled.
 *
 * 26 May 2002 Matthias Benkmann <m.s.b@gmx.net>
 * - set close-on-exec bit for initctl_fd to prevent it from being inherited
 *   by boot scripts and finalprog.
 *
 * 10 Jan 2004 Matthias Benkmann <msb@winterdrache.de>
 * - changed the algorithm to compute the respawn rate in spawn() to something
 *   that at least gets close to computing a real respawn rate 
 * - increased MAX_RESPAWN_RATE to 30, because I think 1 respawn every 3s
 *   is still okay. Otherwise someone could kill off a getty simply by
 *   repeatedly logging in and exiting quickly enough.
 * - fixed bug that caused wrong getty entry to be respawned after quenching
 *   another one
 * - changed quenching of entries that respawn too fast to retry spawning
 *   entry after QUENCH_TIME seconds
 *
 *
 * KNOWN BUGS:
 * - if a directory is specified as bootprog (documented as unsupported in the
 *   manpage), the resulting error message is very obscure. It consists of 
 *   the name of the directory concatenated
 *   to itself until the whole boot_prog buffer is filled. 
 */

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/ioctl.h>
#include <dirent.h>
#include <termios.h>
#include <utmp.h>
#include <sched.h>
#include <time.h>
#include "defines.h"
#include "pathnames.h"
#include "linux_reboot.h"
#include "xstrncpy.h"
#include "nls.h"
#include "simpleinit.h"

#define MAXPARAMS  128  /* max num of args you can use with ctrlaltdel prog */
#define CMDSIZ     150	/* max size of a line in inittab */
#define NUMCMD     30	/* max number of lines in inittab */
#define NUMTOK     20	/* max number of tokens in inittab command */
#define PATH_SIZE  (CMDSIZ+CMDSIZ+1)

#define MAX_RESPAWN_RATE  30 /*  number of respawns per 100 seconds  */
#define RESPAWN_DELAY_INIT 100000 /* initial value of the avg. respawn delay */
#define QUENCH_TIME 600 /* if respawning too fast, disable for ... seconds */

#define TZFILE "/etc/TZ"
char tzone[CMDSIZ];
/* #define DEBUGGING */

/* define this if you want to watch init look for the script when not */
/* passed a complete pathname */
/* #define DEBUG_GETPATH */

/* Define this if you want init to ignore the termcap field in inittab for
   console ttys. */
/* #define SPECIAL_CONSOLE_TERM */

/* define this to make user space init work relative to $HOME instead of
   the directory where it was started or INIT_ROOT. */
/* #define RELATIVE_TO_HOME */

#define ever (;;)

struct initline {
	pid_t		pid;
	char		tty[10];
	char		termcap[30];
	char		*toks[NUMTOK];
	char		line[CMDSIZ];
	struct timeval	last_start;
	signed long	avg_respawn_delay; /* An approximation of the average 
	                                    delay between 2 respawns
	                                    over the last few respawns (in ms) */
};

struct initline inittab[NUMCMD];
int numcmd;
int stopped = 0;	/* are we stopped */
static char boot_prog[PATH_SIZE] = _PATH_RC;
static char init_root[PATH_SIZE] = "\0";
static char script_prefix[PATH_SIZE] = "\0";
static char final_prog[PATH_SIZE] = "\0";
static char init_path[PATH_SIZE] = "\0";
static char cad_prog[PATH_SIZE] = "\0";
static char _path_utmp[PATH_SIZE] = _PATH_UTMP;
static char _path_wtmp[PATH_SIZE] = _PATH_WTMP;
static char _path_inittab[PATH_SIZE] = _PATH_INITTAB;
static char initctl_name[PATH_SIZE] = "/dev/initctl";
static char pidfile_name[PATH_SIZE] = "/dev/init.pid";
static int quiet_rollback = 0;
static int caught_sigint = 0;
static int sigint_critical = 1; 
static int have_cad_prog = 0;
static int userspace = 0;  /*1 means init has been started by user not kernel*/
static int initctl_fd = -1;
static sigset_t all_signals,no_signals,oldsigmask;
static int sigenum;

/* 0: no stop script runs, >0: pid of running stop script */
static pid_t rb_victim_pid = 0; 
static int rb_forced = 0; 
static pid_t rb_watchdog_pid = 0;
static pid_t rb_requester_pid = 0; /*pid of process that requested rollback*/
/*
NULL: rollback all, 
RB_SCRIPT_CLEARED: rollback target cleared by COMMAND_CLEAR
RB_NOROLLBACK: no rollback in progress
*/
#define RB_SCRIPT_CLEARED ((struct script_struct*)&rb_requester_pid)
#define RB_NOROLLBACK ((struct script_struct*)&rb_victim_pid)
struct script_struct* rb_script=RB_NOROLLBACK; 

void err (const char *s);
int do_rc_tty (const char *path);
int run_file (const char *path);
int str2bool(const char* str);
char* fix_parameter(char* cline, const char* id);
void fix_userspace_paths(void);
int exec_with_params(const char* path,const char* preargs,const char* postargs);
int split_param(char** str,int* argc,char* argv[],int argc_max);
void spawn (int i), read_inittab (void);
void hup_handler (int sig);
void sigtstp_handler (int sig);
void int_handler (int sig);
void sigchild_handler (int sig);
void sigquit_handler (int sig);
void userspace_sigquit_handler (int sig);
void sigterm_handler (int sig);
#ifdef SET_TZ
void set_tz (void);
#endif
void write_wtmp (void);
void cleanup_utmp(void);
void utmp_init_process(int id);
pid_t mywait (int *status);
int run_command (const char *file, const char *name, pid_t pid,struct script_struct* waiting_script);
void ystrncat(char* dest,const char* src,ssize_t n);

/* like strcat, but dest will never contain more than n characters after
the call (i.e. a 0-terminator will be found among the first n characters,
unless n<=0)
NOTE: This is different from strncat where n is the maximum number of
characters appended to dest. Another difference is that n is ssize_t and
negative numbers are treated as 0.*/
void ystrncat(char* dest,const char* src,ssize_t n)
{
  while(n>1 && *dest!=0) {++dest; --n;}
  while(n>1 && *src!=0) {*dest++=*src++; --n;};
  if (n>0) *dest=0;
}

#define ystrcat(dest,src) ystrncat((dest),(src),sizeof(dest))
#define ystrcpy(dest,src) xstrncpy((dest),(src),sizeof(dest))

/* like sprintf but protects against overflow and works if source(s) and
target overlap*/
#define ysprintf(dest,rest...) \
{ \
  char buffy[sizeof(dest)]; \
  snprintf(buffy,sizeof(buffy), ##rest); \
  buffy[sizeof(dest)-1]=0; \
  memcpy((dest),buffy,sizeof(dest)); \
}

void err (const char *s)
{
	int fd;
	
	if (userspace) fd=STDOUT_FILENO; 
	else if((fd = open("/dev/console", O_WRONLY)) < 0) return;

	write(fd, "init: ", 6);	
	write(fd, s, strlen(s));
	if (!userspace) close(fd);
}

int main(int argc, char *argv[])
{
	int			vec,i;
	pid_t			pid;
	struct sigaction	sa;

	if (getpid()!=1) 
	{
	  printf("init: started in user space, abort with Ctrl-\\ (SIGQUIT)\n");
	  userspace=1;
	  setenv("USERSPACE_INIT","1",1);
	}

	sigfillset(&all_signals);
	sigemptyset(&no_signals);

#ifdef SET_TZ
	set_tz();
#endif
	sigfillset (&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler= int_handler;
	sigaction (SIGINT, &sa, NULL);
	sa.sa_handler = sigchild_handler;
	sigaction (SIGCHLD, &sa, NULL);
	sa.sa_handler = userspace_sigquit_handler;
	sigaction (SIGQUIT, &sa, NULL);
	if (userspace) fix_userspace_paths(); else
	{
  	  sa.sa_handler = sigtstp_handler;
	  sigaction (SIGTSTP, &sa, NULL);
	  sa.sa_handler = sigterm_handler;
	  sigaction (SIGTERM, &sa, NULL);
	  sa.sa_handler = sigquit_handler;
	  sigaction (SIGQUIT, &sa, NULL);

	  my_reboot (LINUX_REBOOT_CMD_CAD_OFF);
	} 

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
	
	/*  Find script to run. Command-line overrides config file overrides
	    built-in default  */
	for (i = 0; i < NUMCMD; i++) inittab[i].pid = -1;
	read_inittab ();
	
	/* command line arguments replace inittab arguments to bootprog */
	if (argc>1) 
	{
	  ssize_t ofs=strcspn(boot_prog," \t\r\n");
	  if (ofs>0) boot_prog[ofs]=0;
	}
	/* if an executable name is on the command line, replace bootprog,
	otherwise append command line arguments */
	for (i = 1; i < argc; i++) {
			char path[PATH_SIZE];
			
			ystrcat(boot_prog," "); /* Yes, I'm ashamed of myself for writing this line ;-) */
			ystrcat(boot_prog,argv[i]);
			strcpy (path, script_prefix);
			strcat (path, argv[i]);
			if (access (path, R_OK | X_OK) == 0)
			{
			  struct stat Stat;
			  if ((stat(path,&Stat)==0) && (!S_ISDIR(Stat.st_mode)))
			  	strcpy (boot_prog, path);
			}	
	}

	if ( ( initctl_fd = open (initctl_name, O_RDWR|O_NONBLOCK, 0) ) < 0 ) {
		mkfifo (initctl_name, S_IRUSR | S_IWUSR);
		if ( ( initctl_fd = open (initctl_name, O_RDWR|O_NONBLOCK, 0) ) < 0 )
			err ( _("error opening fifo\n") );
	}
	fcntl(initctl_fd,F_SETFD,FD_CLOEXEC); /*make sure no one inherits the fd*/
	
	if (userspace){
	  pid_t mypid=getpid();
	  int fd;
	  if (((fd=open(pidfile_name,O_WRONLY|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR)) < 0) ||
	      (write(fd,&mypid,sizeof(mypid)) < (ssize_t)sizeof(mypid) )
	     ) {err( _("error writing init.pid\n") ); exit(1); }
	  close(fd);   
	}
	
	if (!userspace)
	  for (i = 0; i < getdtablesize (); i++)
		if (i != initctl_fd) close (i);

	/*If we get a SIGTSTP before multi-user mode, do nothing*/
	while(stopped)	
		pause();

	if ( do_rc_tty (boot_prog) ) err ( _("Error executing boot program\n"));
	sigint_critical=0;

	while(stopped)	/* if we get SIGTSTP while executing boot scripts */
		mywait(&vec);

	write_wtmp();	/* write boottime record */

	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler= hup_handler;
	sigaction (SIGHUP, &sa, NULL);

	cleanup_utmp();

	for(i = 0; i < numcmd; i++)
		spawn(i);
	
	if (final_prog[0] != '\0') {
		sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
		switch ( fork () )
		{
		  case 0:   /*  Child   */
		    for (sigenum=1; sigenum<NSIG; ++sigenum) signal(sigenum, SIG_DFL);
		    sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
		    execl (final_prog, final_prog, "start", NULL);
		    err ( _("error running finalprog\n") );
		    _exit (1);
		    break;
		  case -1:  /*  Error   */
		    sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
		    err ( _("error forking finalprog\n") );
		    break;
		  default:  /*  Parent  */
		    sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
		    break;
		}
	}

	for ever {
		pid = mywait (&vec);
		if (pid < 1) continue;

		/* clear utmp entry, and append to wtmp if possible */
		{
		    struct utmp *ut;

		    utmpname(_path_utmp);
		    setutent();
		    while((ut = getutent())) {
			if(ut->ut_pid == pid) {
			    time(&ut->ut_time);
			    memset(&ut->ut_user, 0, UT_NAMESIZE);
			    memset(&ut->ut_host, 0, sizeof(ut->ut_host));
			    ut->ut_type = DEAD_PROCESS;
			    ut->ut_pid = 0;
			    ut->ut_addr = 0;
			    pututline(ut);
#ifdef HAVE_updwtmp
			    updwtmp(_path_wtmp,ut);
#endif
			    break;
			}
		    }
		    endutent();
		}
		for(i = 0; i < numcmd; i++) {
			if(pid == inittab[i].pid || inittab[i].pid < 0) {
				if(stopped) 
				  inittab[i].pid = -1;
				else 
				  spawn(i);
			}
		}
	}
}	

#define MAXTRIES 3 /* number of tries allowed when giving the password */

/*
 * run boot script(s). The environment is passed to the script(s), so the RC
 * environment variable can be used to decide what to do.
 * RC may be set from LILO.
 * [RETURNS] 0 on success (exit status convention), otherwise error.
 */
int do_rc_tty (const char *path)
{
    int status;
    pid_t pid, child=0; /* =0 to silence compiler warning*/
    sigset_t ss;

    if (caught_sigint) return 0;
    
    /*  Launch off a subprocess to start a new session (required for frobbing
	the TTY) and capture control-C  */

    if (!userspace){ 
      sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
      switch ( child = fork () )
      {
        case 0:   /*  Child  */
  		for (status = 1; status < NSIG; status++) signal (status, SIG_DFL);
  		sigfillset (&ss);
  		sigprocmask (SIG_UNBLOCK, &ss, NULL);
  		sigdelset (&ss, SIGINT);
  		sigdelset (&ss, SIGQUIT);
  		setsid ();
  		ioctl (0, TIOCSCTTY, 0);  /*  I want my control-C  */
  		while(1) sigsuspend (&ss);  /*  Should never return, should just be killed  */
  		break;             /*  No-one else is controlled by this TTY now   */
        case -1:  /*  Error  */
  		  sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
  		  return (1);
  		/*break;*/
        default:  /*  Parent  */
        	sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
  		break;
      }
    }  
    /*  Parent  */
    run_file (path);
    while (1)
    {
	pid = mywait (&status);
	if ( !userspace && (pid == child) )
	    return (WTERMSIG (status) == SIGINT) ? 0 : 1;
	if (pid < 0) break;
    }
    if (!userspace){ 
       kill (child, SIGKILL);
       while (waitpid (child, NULL, 0) != child)  /*  Nothing  */;
    }  
    return 0;
}   /*  End Function do_rc_tty  */

int run_file (const char *path)
{
    const char *ptr;
    ssize_t len;
    char service_name[256];

    if ( ( ptr = strrchr ( (char *) path, '/' ) ) == NULL ) ptr = path;
    else ++ptr;
    len=strcspn(ptr," \t\r\n");
    if (len>=256-1) len=256-1;
    memcpy(service_name,ptr,len);
    service_name[len]=0;
    return (run_command (path, service_name, 0, NULL) == SIG_FAILED) ? 1 : 0;
}   /*  End Function run_file  */

void spawn (int i)
{
	pid_t pid;
	int j;
	signed long ds_taken;
	struct timeval ct;
	int rate;
	
	if (userspace) return;	
	if (inittab[i].toks[0] == NULL) return;
	
	/*  Check if respawning too fast  */
	gettimeofday (&ct, NULL);
	ds_taken = ct.tv_sec - inittab[i].last_start.tv_sec;
	if (ds_taken<0) ds_taken=0; /* not supposed to happen */

	if (inittab[i].avg_respawn_delay<0) /* entry has been quenched */
	{
	  if (ds_taken > QUENCH_TIME)
	    inittab[i].avg_respawn_delay=RESPAWN_DELAY_INIT;
	  else   
	    return;
	}

	/* Ignore anything more than 100s since last respawn, to avoid
	pulling down the respawn rate too much. */
	if (ds_taken>100) ds_taken=100;
	
	ds_taken *= 1000; /* convert to ms */
	
	inittab[i].avg_respawn_delay = (inittab[i].avg_respawn_delay + ds_taken) / 2;
	
	rate=(100000+1)/(1+inittab[i].avg_respawn_delay); /* 1+ to avoid div by 0*/
	if (rate > MAX_RESPAWN_RATE) {
		char txt[256];

		inittab[i].pid = -1;
		inittab[i].avg_respawn_delay = -1;
		inittab[i].last_start = ct;
		sprintf (txt,"respawning: \"%s\" too fast: quenching entry\n",
			 inittab[i].tty);
		err (_(txt));
		return;
	}

	sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
	if((pid = fork()) < 0) {
		sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
		inittab[i].pid = -1;
		err(_("fork failed\n"));
		return;
	}
	if(pid) {
		/* this is the parent */
		sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
		inittab[i].pid = pid;
		inittab[i].last_start = ct;
		sched_yield ();
		return;
	} else {
		/* this is the child */
		char term[40];
#ifdef SET_TZ
		char tz[CMDSIZ];
#endif
		char *env[3];
		
		for (sigenum=1; sigenum<NSIG; ++sigenum) signal(sigenum, SIG_DFL);
		setsid();
		
		utmp_init_process(i); 
		
		for(j = 0; j < getdtablesize(); j++)
			(void) close(j);

		sprintf(term, "TERM=%s", inittab[i].termcap);
		env[0] = term;
		env[1] = NULL;
#ifdef SET_TZ
		sprintf(tz, "TZ=%s", tzone);
		env[1] = tz;
#endif
		env[2] = NULL;
		
		sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
		execve(inittab[i].toks[0], inittab[i].toks, env);
		err(_("exec failed\n"));
		sleep(5);
		_exit(1);
	}
}

/*returns 1 if str is "yes", "true" or "1" (case insensitive), otherwise 0*/
int str2bool(const char* str)
{
  return 
    (!strcasecmp(str,"yes")|| !strcasecmp(str,"true")|| !strcasecmp(str,"1"));
}

/*tests if cline starts with an arbitrary number of whitespace, followed by id 
(case insensitive!!), followed by an arbitrary
number of = signs or whitespace. If so, then trailing whitespace is removed.
RETURN value:
 NULL if cline doesn't match id,
 pointer to processed string (i.e. component after =) otherwise
ATTENTION! The processed string may be empty (i.e. =="\0") if nothing but
ws or = signs followed id.
*/
char* fix_parameter(char* cline, const char* id)
{            /*written by MSB*/
  char* str;
  int len=strlen(id);

  while (isspace(*cline)) ++cline;
  if (strncasecmp (cline,id,len)!=0) return NULL;
  if ((cline[len]!='=') && (!isspace(cline[len]))) return NULL; 
 
  for (cline+=len; isspace(*cline) || (*cline=='='); ++cline);
  if (*cline=='\0') return cline;
  
  /*cline is not empty ==> cut off trailing whitespace*/
  for (str=cline; *str!=0; ++str);
  for (--str; ((str!=cline) && isspace(*str)); --str);
  if (!isspace(*str)) ++str;
  *str=0;
  return cline;
}

void read_inittab (void)
{
	FILE *f;
	char buf[CMDSIZ];
	int i,j,k;
	int has_prog = 0;
	char *ptr, *getty;
	char prog[PATH_SIZE];
#ifdef SPECIAL_CONSOLE_TERM
	char tty[50];
	struct stat stb;
#endif
	char *termenv;
	
	termenv = getenv("TERM");	/* set by kernel */
	/* termenv = "vt100"; */
			
	if(!(f = fopen(_path_inittab, "r"))) {
		err(_("cannot open inittab\n"));
		return;
	}

	prog[0] = '\0';
	i = 0;
	while(!feof(f) && i < NUMCMD - 2) {
		if(fgets(buf, CMDSIZ - 1, f) == 0) break;
		buf[CMDSIZ-1] = 0;

		for(k = 0; k < CMDSIZ && buf[k]; k++)
			if ((buf[k] == '#') || (buf[k] == '\n')) break;
		buf[k]=0;
		/*treat lines that start with whitespace as empty*/
		for(--k; k>=0 && isspace(buf[k]) ;--k) buf[k]=0;

		if(buf[0] == 0) continue; /*skip empty lines*/

		if ((ptr=fix_parameter(buf, "bootprog"))!=NULL) {
			strcpy (prog, ptr);
			has_prog = 1;
			continue;
		} else
		if ((ptr=fix_parameter(buf, "fileprefix"))!=NULL) {
			size_t len;
			ystrcpy(script_prefix,init_root);
			len=strlen(script_prefix);
			xstrncpy (script_prefix+len, ptr,PATH_SIZE-len);
			setenv ("INIT_D", script_prefix, 1);
			continue;
		} else
		if ((ptr=fix_parameter(buf, "PATH"))!=NULL) {
			setenv ("PATH", ptr, 1);
			continue;
		} else
		if ((ptr=fix_parameter(buf, "INIT_PATH"))!=NULL) {
			xstrncpy (init_path, ptr, sizeof(init_path));
			continue;
		} else
		if ((ptr=fix_parameter(buf, "finalprog"))!=NULL) {
			strcpy (final_prog, ptr);
			continue;
		} else
		if ((ptr=fix_parameter(buf, "quietrollback"))!=NULL) {
			quiet_rollback=str2bool(ptr);
			continue;
		} else
		if ((ptr=fix_parameter(buf, "ctrlaltdel"))!=NULL) {
			strcpy(cad_prog,ptr);
			have_cad_prog=1;
			continue;
		}
		
		if (strchr(buf,':')==NULL) {
		  err(_("unrecognized command in inittab\n"));
		  continue;
		} 
		
		strcpy(inittab[i].line, buf);
		inittab[i].avg_respawn_delay=RESPAWN_DELAY_INIT;
		strtok(inittab[i].line, ":");
		xstrncpy(inittab[i].tty, inittab[i].line, 10);
		xstrncpy(inittab[i].termcap, strtok(NULL, ":"), 30);

		getty = strtok(NULL, ":");
		strtok(getty, " \t\n");
		inittab[i].toks[0] = getty;
		j = 1;
		while((ptr = strtok(NULL, " \t\n")))
			inittab[i].toks[j++] = ptr;
		inittab[i].toks[j] = NULL;

#ifdef SPECIAL_CONSOLE_TERM
		/* special-case termcap for the console ttys */
		(void) sprintf(tty, "/dev/%s", inittab[i].tty);
		if(!termenv || stat(tty, &stb) < 0) {
			err(_("no TERM or cannot stat tty\n"));
		} else {
			/* is it a console tty? */
			if(major(stb.st_rdev) == 4 && minor(stb.st_rdev) < 64) 
				xstrncpy(inittab[i].termcap, termenv, 30);
		}
#endif

		i++;
	}
	fclose(f);
	numcmd = i;
	if (has_prog) {
 		ystrcpy (boot_prog, script_prefix);
		ystrcat (boot_prog, prog);
	}
}

void hup_handler (int sig)
{
	int i,j;
	int oldnum;
	struct initline	savetab[NUMCMD];
	int had_already;

	memcpy(savetab, inittab, NUMCMD * sizeof(struct initline));
	oldnum = numcmd;		
	read_inittab ();
	
	for(i = 0; i < numcmd; i++) {
		had_already = 0;
		for(j = 0; j < oldnum; j++) {
			if(!strcmp(savetab[j].tty, inittab[i].tty)) {
				had_already = 1;
				if((inittab[i].pid = savetab[j].pid) < 0)
					spawn(i);
			}
		}
		if(!had_already) spawn(i);
	}
	sig=sig;
}

void sigtstp_handler (int sig)
{
    stopped = ~stopped;
    if (!stopped) hup_handler (sig);
}   /*  End Function sigtstp_handler  */

void sigterm_handler (int sig)
{
    int i;
    for (i = 0; i < numcmd; i++)
	if (inittab[i].pid > 0) kill (inittab[i].pid, SIGTERM);
    sig=sig;
}   /*  End Function sigterm_handler  */

void int_handler (int sig)
{
	pid_t pid;
	static time_t last_press=0;
	time_t t;
	sigset_t oldmask;
	int s;
		
        /*after running all rc scripts, sigint is not critical anymore as
        it just tells us we need to run reboot (which does *not* mean that
        the system is really going down, since reboot could be just a program
        that gives a message). For that reason behaviour of init should not
        change based on caught_sigint. */
	caught_sigint = sigint_critical;
	sync();
	sync();
	
	/*need at least 5 secs between CAD presses. Every keypress resets
	the countdown so that you can hammer CAD freely for a prolonged 
	period of time without adverse side effects.*/
	t=last_press;
	last_press=time(NULL);
	if (last_press-t<5) return;
	
	sigprocmask(SIG_BLOCK,&all_signals,&oldmask);
	pid = fork();
	if (pid == 0) {	/* reboot properly... */
		for (s=1; s<NSIG; ++s) signal(s, SIG_DFL);
		sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
		if (have_cad_prog) exec_with_params(cad_prog,NULL,NULL);
		else 
		if (!userspace) execl(_PATH_REBOOT, _PATH_REBOOT, NULL);
	} 
	else sigprocmask(SIG_SETMASK,&oldmask,NULL);
	
	if (pid > 0) return; /*parent returns if fork() was successful*/

	/* fork or exec failed, try the hard way unless either in user space
	or the user specified his own program (in which case he probably
	doesn't like the hard reboot) */
	if (!(userspace || have_cad_prog))
		my_reboot(LINUX_REBOOT_CMD_RESTART);
	if (pid==0) exit(1);
	sig=sig;
}

void sigchild_handler (int sig)
{
/* empty handler to make sure SIGCHLD doesn't get ignored */
  sig=sig;
}

void sigquit_handler (int sig)
{
    execl (_PATH_SHUTDOWN, _PATH_SHUTDOWN, NULL); /*  It knows pid=1 must sleep  */
    err("could not exec shutdown\n");
    sig=sig;
}

void userspace_sigquit_handler (int sig)
{
    exit(0);
    sig=sig;
}

#ifdef SET_TZ
void set_tz (void)
{
	FILE *f;
	int len;

	if((f=fopen(TZFILE, "r")) == (FILE *)NULL) return;
	fgets(tzone, CMDSIZ-2, f);
	fclose(f);
	if((len=strlen(tzone)) < 2) return;
	tzone[len-1] = 0; /* get rid of the '\n' */
	setenv("TZ", tzone, 0);
}
#endif

void write_wtmp (void)
{
    struct utmp ut;
    
    memset((char *)&ut, 0, sizeof(ut));
    strcpy(ut.ut_line, "~");
    memset(ut.ut_name, 0, sizeof(ut.ut_name));
    time(&ut.ut_time);
    ut.ut_type = BOOT_TIME;
#ifdef HAVE_updwtmp
    updwtmp(_path_wtmp,&ut);
#endif
}     

/*walks through utmp and sets all entries with ut_type == *_PROCESS to 
DEAD_PROCESS. All ut_pid fields will be set to 0.*/
void cleanup_utmp (void)
{ /*written by MSB*/
  struct utmp *ut;

  utmpname(_path_utmp);
  setutent();
  while((ut = getutent())) 
  {
    /*DEAD_PROCESS is included here to make sure everything is properly 
    cleared in case the DEAD_PROCESS entry was created by someone else*/
    if ((ut->ut_type == INIT_PROCESS) ||
        (ut->ut_type == DEAD_PROCESS) ||
        (ut->ut_type == USER_PROCESS) ||
        (ut->ut_type == LOGIN_PROCESS))
    {
      memset(&ut->ut_user, 0, UT_NAMESIZE);
      memset(&ut->ut_host, 0, sizeof(ut->ut_host));
      ut->ut_type = DEAD_PROCESS;
      ut->ut_pid = 0;
      ut->ut_addr = 0;
      if (pututline(ut)==NULL) break;
    } else
    if (ut->ut_pid != 0) { ut->ut_pid=0; if (pututline(ut)==NULL) break; }
  }
  endutent();
}

/*creates an INIT_PROCESS entry in utmp. id (converted to a string) will be 
used as ut_id.
The first existing utmp line with that ut_id will be recycled if such a line
exists (pututline()'s behaviour)*/
void utmp_init_process(int id)
{
  struct utmp ut;
  utmpname(_path_utmp);
  setutent();
  memset(&ut,0,sizeof(ut));
  ut.ut_type=INIT_PROCESS;
  ut.ut_pid=getpid();
  
  ut.ut_id[3]=id%10+'0';
  ut.ut_id[2]=(id/10)%10+'0';
  ut.ut_id[1]=(id/100)%10+'0';
  ut.ut_id[0]=(id/1000)%10+'0';
  
  time(&ut.ut_time);
  pututline(&ut);
  endutent();
}

struct needer_struct
{
    struct needer_struct *next;
    pid_t pid;
};

struct service_struct
{
    struct service_struct *prev, *next;    /*  Script services chain         */
    struct needer_struct *needers;         /*  Needers waiting for service   */
    struct script_struct *attempting_providers;
    int failed;                /*  TRUE if attempting provider failed badly  */
    char name[1];/* 1 char for zero terminator */
};

struct script_struct
{
    pid_t pid; /* pid of script itself */
    pid_t blocker_pid; /* pid of blocking need/provide (only defined if waiting_for != NULL) */
    struct script_struct *prev, *next;              /*  For the list         */
    struct service_struct *first_service, *last_service; /*First is true name*/
    struct script_struct *next_attempting_provider; /*  Provider chain       */
    struct service_struct *waiting_for;  /* script blocks waiting for this  */
};
 
struct list_head
{
    struct script_struct *first, *last;
    unsigned int num_entries;
};

/*entries are appended to the front, i.e. the first entry is the most recent*/
static struct list_head available_list = {NULL, NULL, 0};
static struct list_head starting_list = {NULL, NULL, 0};
static struct service_struct *unavailable_services = NULL;  /*  For needers  */

/* The pointer orgy used for service accounting is very complex. It took me
a while to grasp it. The following will hopefully help other programmers 
understand it faster. -- MSB

available_list--> SCRIPT <--> SCRIPT <--> SCRIPT <--> SCRIPT
                   |            |           |           |           
                   v            v           v           v
                 SERVICE*     SERVICE*   SERVICE*     SERVICE*  
                   ^
                   |
                   v
                 SERVICE
                   ^
                   |                 * The first service (every script has 
                   v                   at least this one) is always the real
                 SERVICE               name of the script.


      # waiting_for ------->SERVICE*    SERVICE*    SERVICE*
                    |          ^           ^           ^
                    |          |           |           | 
starting_list--> SCRIPT <--> SCRIPT <--> SCRIPT <--> SCRIPT
                   |                      ^   |         ^\------------------
                   v                      |   |                            |
                 SERVICE*                 |   |                            |
                   ^                      |    \-next_attempting_provider /
                   |                      |  
                   v                      |
               @ SERVICE attempting_providers
                   ^
                   |
                   v
                 SERVICE needers--> NEEDER --> NEEDER --> NEEDER
                 
@ A service is always in the list of the provider that is currently
running. The providers in the linked list defined by attempting_providers and
next_attempting_provider are currently blocked.

# Only starting_list has scripts with non-NULL waiting_for entries.

unavailable_services --> SERVICE <--> SERVICE <--> SERVICE
                            |
                            v
                          NEEDER   
                            |
                            v
                          NEEDER   
                             
Note: Services in the unavailable list do not have waiting needers. This
change was introduced with the deadlock detection. However this might
change again in future. Any code dealing with unavailable_list should
deal with needers properly. 

*/   

static void wake_rb_watchdog(void);
static void kill_rb_watchdog(void);
static void finish_rollback(int sig);
static void cleanup_and_continue_rollback(void);
static int process_pidstat (pid_t pid, int status);
static void process_command (unsigned command,const struct command_struct *comdata);
static struct service_struct *find_service_in_list (const char *name,
						    struct service_struct *sv);
static struct script_struct *find_script_byname
    (const char *name,struct list_head *head, struct service_struct **service);
static struct script_struct *find_script_bypid (pid_t pid,
						struct list_head *head);
static struct script_struct *find_script_byservice 
		(const struct service_struct* service, struct list_head *head);
static void insert_entry (struct list_head *head, struct script_struct *entry);
static void remove_entry (struct list_head *head, struct script_struct *entry);
static void signal_needers (struct service_struct *service, int sig);
static void handle_nonworking (struct script_struct *script);
static void show_scripts (FILE *fp, const struct script_struct *script,
			  const char *type);
static const char *get_path (const char *file);
static void clean_up_waiting_for(const struct service_struct* service);
static int waits_for(	const struct script_struct* who, 
			const struct script_struct* forwhat);
static void output_deadlock_scripts(	const struct script_struct* start, 
					const struct script_struct* stop,
					const struct service_struct* service);

pid_t mywait (int *status)
/*  [RETURNS] The pid for a process to be reaped, 0 if no process is to be
    reaped, and less than 0 if the boot scripts appear to have finished.
*/
{
    char buffer[COMMAND_SIZE];
    pid_t pid;
    int ival;
    sigset_t oldss, suspss;
    struct command_struct *command = (struct command_struct *) buffer;

    if (initctl_fd < 0) return waitpid (-1, status, 0);
    if (status == NULL) status = &ival;
    
    sigprocmask(SIG_SETMASK,NULL,&oldss);
    suspss=oldss;
    sigdelset(&suspss,SIGCHLD);
    sigdelset(&suspss,SIGQUIT);
    
    while(1)
    { 
      sigprocmask(SIG_BLOCK,&all_signals,NULL);
      if ( ( pid = waitpid (-1, status, WNOHANG) ) > 0 )
      {
   	sigprocmask(SIG_SETMASK,&oldss,NULL);
   	return process_pidstat (pid, *status);
      }
      
      if ((read(initctl_fd, buffer, COMMAND_SIZE))==COMMAND_SIZE)
      {
        sigprocmask(SIG_SETMASK,&oldss,NULL);
        process_command(command->command,command);
        return 0;
      }
      sigsuspend(&suspss);
    }  
}   /*  End Function mywait  */

static pid_t process_pidstat (pid_t pid, int status)
/*  [RETURNS] The pid for a process to be reaped, 0 if no process is to be
    reaped, and less than 0 if the boot scripts appear to have finished.
*/
{
    int failed;
    struct script_struct *script;
    struct service_struct *service;
    
    if (pid<=0) return 0;

    /* handle watchdog */
    if (pid==rb_watchdog_pid){
    	rb_watchdog_pid=0;
    	return 0;
    }
   
    /* handle stop script */
    if (pid == rb_victim_pid)
    {
     	char txt[256];
	
	rb_victim_pid=0;
	kill_rb_watchdog();
	
  	/* test if stop script successful or forced rollback */
  	if ( (rb_forced) || 
  	     ( WIFEXITED (status) && (WEXITSTATUS (status) == 0) ) )
  	  	cleanup_and_continue_rollback();
  	else
  	{  /* stop script failed */
  	    struct script_struct* victim = available_list.first;
  	    if (victim != NULL) /* COMMAND_CLEAR might have killed it */
  	    {
  	    	snprintf (txt,255, "Stopping service failed: %s\n",
  			     victim->first_service->name);
  	    	if (!quiet_rollback) err (txt);
  	    }	
	    finish_rollback(SIG_NOT_STOPPED);
  	}
      	return 0;
    } 

    /* handle program unrelated to init */
    if ( ( script = find_script_bypid (pid, &starting_list) ) == NULL )
	return pid;
	
    /* handle start script */
    remove_entry (&starting_list, script);
    if ( WIFEXITED (status) && (WEXITSTATUS (status) == 0) )
    {
	struct script_struct *provider;
	struct script_struct *next_attempting_provider=NULL;
	
	/*  Notify needers and other providers. */
	for (service = script->first_service; service != NULL;
	     service = service->next)
	{
	    signal_needers (service, SIG_PRESENT);
	    for (provider = service->attempting_providers; provider != NULL;
		 provider = next_attempting_provider)
		 {
		   next_attempting_provider=provider->next_attempting_provider;
		   provider->next_attempting_provider=NULL;
		   kill (provider->blocker_pid, SIG_PRESENT);
		 }  
	    service->attempting_providers = NULL;
	    clean_up_waiting_for(service);
	}
	insert_entry (&available_list, script);
	script->waiting_for=NULL; /*in case script terminates before need */
    } else
    {
        failed = ( WIFEXITED (status) && (WEXITSTATUS (status) == 2) ) ? 0 : 1;
        for (service = script->first_service; service != NULL;
	     service = service->next)
  	    	service->failed = failed;
    	handle_nonworking (script);
    }
    return (starting_list.num_entries < 1) ? -1 : 0;
}   /*  End Function process_pidstat  */

static void cleanup_and_continue_rollback(void)
{  
    char txt[256];
    struct service_struct *service, *next_service;
    struct script_struct* victim = available_list.first;
	    
    if (victim != NULL) /* COMMAND_CLEAR might have killed it */
    {
    	snprintf (txt,255, "Stopped service: %s\n",
		     victim->first_service->name);
   	remove_entry (&available_list, victim);
 	for (service = victim->first_service; service != NULL;
    	     service = next_service)
    	     {
    	       	next_service = service->next;
    	       	free(service);
    	     }     
    	            		
    	free (victim);
    	if (!quiet_rollback) err (txt);
    }	
    process_command(COMMAND_ROLLBACK,NULL); /* continue rollback */
}

static void finish_rollback(int sig)
{
	kill(rb_requester_pid,sig); 
  	rb_requester_pid=0;
  	rb_script=RB_NOROLLBACK;
  	rb_forced=0;
}

static void process_command (unsigned command,const struct command_struct *comdata)
{
    int ival;
    struct script_struct *script,*script2;
    struct service_struct *service;
    unsigned command_flags;

    command_flags=command & ~CLEAR_FLAGS;
    command &= CLEAR_FLAGS;

    /* when adding new commands, don't forget that they might be called 
    while a rollback is in progress (i.e. rb_script != RB_NOROLLBACK). 
    You should kill the requester with
    SIG_ROLLBACK in that case (unless the function is unaffected like
    COMMAND_TEST and COMMAND_DUMP_LIST)*/
    
    switch (command)
    {
      case COMMAND_TEST:
        if (find_script_byname (comdata->name, &available_list, NULL) != NULL)
        	kill(comdata->pid,SIG_PRESENT); else
        if ( ( service = find_service_in_list (comdata->name,
        				unavailable_services) ) != NULL )
	{
		if (service->failed) 
			kill(comdata->pid,SIG_FAILED); 
		else	kill(comdata->pid,SIG_NOT_PRESENT); 
	} 
	else kill (comdata->pid, SIG_NO_INFO);
	break;
      case COMMAND_NEED:
      	if (rb_script != RB_NOROLLBACK)
      	{
      	   if (find_script_byname (comdata->name, &available_list, NULL) != NULL)
	    	kill(comdata->pid,SIG_PRESENT);
	   else kill(comdata->pid,SIG_ROLLBACK);
	   break;
      	}
      	
      	script = find_script_bypid (comdata->ppid, &starting_list);
      	script2= find_script_byname(comdata->name, &starting_list, &service);
      	
      	/* handle case that script needs itself */
      	if ((script!=NULL) && (script==script2)) 
      	{
      	    kill(comdata->pid, SIG_PRESENT);
      	    break;
      	}
      	
      	/* if script2 waits for script1 => abort to prevent deadlock */
      	if (waits_for(script2,script)) 
      	{
      	    output_deadlock_scripts(script2,script,service);
      	    kill(comdata->pid, SIG_FAILED);
      	    break;
      	}
      	
	ival = run_command (comdata->name, comdata->name, comdata->pid, script);
	if (ival != 0) kill (comdata->pid, ival);
	break;
      case COMMAND_CLEAR:
        if ( ( service = find_service_in_list (comdata->name,
				       unavailable_services) ) != NULL )
	{   
	    struct needer_struct *needer, *next_needer;
	    for (needer=service->needers; needer!=NULL; needer=next_needer)
	      {next_needer=needer->next; free(needer); }
	    if (service->prev == NULL) unavailable_services = service->next;
	    else service->prev->next = service->next;
	    if (service->next != NULL) service->next->prev = service->prev;
	    free(service);
	    kill (comdata->pid, SIG_NOT_STOPPED);
	} else
	{
	  if ( ( script = find_script_byname (comdata->name, &available_list,
						&service) ) == NULL )
	  {     
	    kill (comdata->pid, SIG_NOT_PRESENT);
	  } else  
	  {
  	    if ( service->prev == NULL )
	    { /* clear requested with script's real name ==> clear all services*/
              struct service_struct* next_service;

              if (rb_script == script) rb_script=RB_SCRIPT_CLEARED;

	      remove_entry (&available_list, script);
  	      for (service = script->first_service; service != NULL;
     	                 service = next_service)
     	            	{
     	              	  next_service = service->next;
     	              	  free(service);
     	            	}     
	      free (script); 
	    } else
	    {/* clear requested for 'provide'd service ==> clear only that one*/
	      service->prev->next = service->next;
	      if (service->next != NULL) service->next->prev = service->prev;
	      free(service);
	    }
	    kill (comdata->pid, SIG_STOPPED);
	  }  
	}  
        break;
      case COMMAND_ROLLBACK:
        if (comdata != NULL) /* if called with a new rollback target */
        { 
	  /* if this is an all-new rollback, save the pid of the requester */
	  /* and set the rb_forced flag					   */
	  if (rb_script == RB_NOROLLBACK){ 
	  	rb_requester_pid = comdata->pid;
	  	if ((command_flags & FLAG_FORCE)!=0) rb_forced=1;
	  }
	  
	  /* find new rollback target in our structures */
	  /* rollback all? ==> script=NULL */
	  /* rollback target not found? ==> script=RB_NOROLLBACK */
	  /* rollback target found? ==> script=address(rollback_target) */
	  if (comdata->name[0] == '\0') script = NULL;
	  else
          if ( ( script = find_script_byname (comdata->name, 
          				&available_list, NULL) ) == NULL )
			script=RB_NOROLLBACK;
					
	  if (rb_script != RB_NOROLLBACK) /* if rollback already in progress */
	  { 	
		if ( (script != RB_NOROLLBACK) && (script != NULL) &&
		     (rb_script != RB_SCRIPT_CLEARED) )
		{  
		   /* test if new rb target comes after the old one  */
		   /* (i.e. new one requires stopping more services) */
		   struct script_struct* s=script->prev;
		   while ((s != NULL) && (s != rb_script)) s=s->prev;
		   
		   /* if new target is not after old one then ignore it */
		   if (s != rb_script) script=RB_NOROLLBACK;
		}
		
	  	kill (comdata->pid, SIG_ROLLBACK);
	  }
	  
	  /* if a new target has been accepted, update rollback target */	
	  if (script != RB_NOROLLBACK) rb_script=script;
	}  
	
	/* don't continue if we still have a stop script running */
	if (rb_victim_pid > 0) break;
	
	if (rb_script == RB_NOROLLBACK) 
				finish_rollback(SIG_NOT_PRESENT); else
	if (rb_script == RB_SCRIPT_CLEARED) 
				finish_rollback(SIG_NOT_STOPPED); else
	if (rb_script == available_list.first) 
				finish_rollback(SIG_STOPPED); 
	else
	{			/* launch stop script */
	    pid_t pid;
	    struct script_struct *victim = available_list.first;
	    char txt[256];

	    sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
	    if ( ( pid = fork () ) == 0 )   /*  Child   */
	    {
		for (ival = 1; ival < NSIG; ival++) signal (ival, SIG_DFL);
		sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
		if (!userspace){
			open ("/dev/console", O_RDONLY, 0);
			open ("/dev/console", O_RDWR, 0);
			dup2 (1, 2);
		}
		{
                  const char* fpath=get_path(victim->first_service->name);
                  const char* stop="stop";
                  struct stat Stat;
                  
                  /* if service is a directory, then execute boot prog instead */
                  if ((stat(fpath,&Stat)==0) && (S_ISDIR(Stat.st_mode)))
                  {
                    ssize_t ofs=strcspn(boot_prog," \t\r\n");
                    if (ofs>0) boot_prog[ofs]=0;
                    ystrcat(boot_prog," ");
                    ystrcat(boot_prog,fpath);
                    fpath=boot_prog;
                    stop="stop-dir";
                  }
			
		  exec_with_params(fpath, stop, NULL);
		  snprintf(txt,256,"errno %u running program %s\n",
			errno,fpath);
		  err ( txt );
		  _exit (SIG_NOT_STOPPED);
		}  
	    } 
	    else sigprocmask(SIG_SETMASK,&oldsigmask,NULL); 
	    
	    if (pid > 0) {
	      rb_victim_pid=pid;
	      if (rb_forced) wake_rb_watchdog();
	    }
	    	    
  	    if (pid == -1) 
  	    {
  	      snprintf(txt,255,"Could not fork process to stop %s\n",victim->first_service->name);
  	      err(txt);
  	      if (!rb_forced) finish_rollback(SIG_NOT_STOPPED); 
  	      else cleanup_and_continue_rollback();
  	    }
	}
	break;
      case COMMAND_DUMP_LIST:
        sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
	if (fork () == 0) /* Do it in a child process so pid=1 doesn't block */
	{
	    FILE *fp;

	    for (sigenum=1; sigenum<NSIG; ++sigenum) signal(sigenum, SIG_DFL);
	    sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
	    if ( ( fp = fopen (comdata->name, "w") ) == NULL ) _exit (1);
	    show_scripts (fp, available_list.first, "AVAILABLE");
	    show_scripts (fp, starting_list.first, "STARTING");
	    fputs ("UNAVAILABLE SERVICES:\n", fp);
	    for (service = unavailable_services; service != NULL;
		 service = service->next)
		fprintf (fp, "%s (%s)\n", service->name,
			 service->failed ? "FAILED" : "not configured");
	    fclose (fp);
	    _exit (0);
	} else sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
	break;
      case COMMAND_PROVIDE:
	if (rb_script != RB_NOROLLBACK){
		kill(comdata->pid, SIG_ROLLBACK);
		break;
	}
	/*  Sanity check  */
	if ( ( script = find_script_bypid (comdata->ppid, &starting_list) )
	     == NULL )
	{
	    kill (comdata->pid, SIG_NOT_CHILD);
	    break;
	}
	if (find_script_byname (comdata->name, &available_list, NULL) != NULL)
	{
	    kill (comdata->pid, SIG_PRESENT);
	    break;
	}
	script2 = find_script_byname (comdata->name, &starting_list, &service);
	if (script2 == script)
	{ /* provide called again for the same service by the active provider */
	    kill (comdata->pid, SIG_NOT_PRESENT);
	    break;
	} 
	if (script2 != NULL)
	{   /* Someone ELSE is trying to provide, so block unless deadlock  */
  	    /* if script2 waits for script1 => abort to prevent deadlock */
            if (waits_for(script2,script)) 
            {
		output_deadlock_scripts(script2,script,service);
        	kill(comdata->pid, SIG_NOT_CHILD);
        	break;
            }
	    script->next_attempting_provider = service->attempting_providers;
	    service->attempting_providers = script;
	    script->waiting_for=service;
	    script->blocker_pid=comdata->pid;
	    break;
	}
	if ( ( service = find_service_in_list (comdata->name,
					       unavailable_services) )
	     == NULL )
	{   /*  We're the first to try and provide: create it  */
	    if ( ( service =
		   calloc (1, strlen (comdata->name) + sizeof *service) )
		 == NULL )
	    {
		kill (comdata->pid, SIG_NOT_CHILD);
		break;
	    }
	    strcpy (service->name, comdata->name);
	}
	else
	{   /*  Orphaned service: unhook and grab it  */
	    if (service->prev == NULL) unavailable_services = service->next;
	    else service->prev->next = service->next;
	    if (service->next != NULL) service->next->prev = service->prev;
	    service->next = NULL;
	    service->prev = NULL;
	}
	service->prev = script->last_service;
	script->last_service->next = service;
	script->last_service = service;
	kill (comdata->pid, SIG_NOT_PRESENT);
	break;
      default:
	break;
    }
}   /*  End Function process_command  */

static void kill_rb_watchdog(void)
{
  if (rb_watchdog_pid > 0) kill(rb_watchdog_pid,SIGKILL);
  /*don't set rb_watchdog_pid to 0. This is done in process_pidstat()*/
}
 
static void wake_rb_watchdog(void)
{
  sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
  if ((rb_watchdog_pid=fork())==0)
  { /* child */
    for (sigenum=1; sigenum<NSIG; ++sigenum) signal(sigenum, SIG_DFL);
    sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
    sleep(5*60);
    kill(rb_victim_pid,SIGTERM);
    sleep(10);
    kill(rb_victim_pid,SIGKILL);
    _exit(0);
  }
  if (rb_watchdog_pid<0) rb_watchdog_pid=0;
  sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
}

/*this function does not prevent deadlocks, you must check before calling it */
int run_command (const char *file, const char *name, pid_t pid, struct script_struct* waiting_script)
{
    struct script_struct *script;
    struct needer_struct *needer = NULL;
    struct service_struct *service;

    if (find_script_byname (name, &available_list, NULL) != NULL)
	return SIG_PRESENT;
    if (pid != 0)
    {
	needer = calloc (1, sizeof *needer);
	if (needer == NULL) return SIG_FAILED;
	needer->pid = pid;
    }
    script = find_script_byname (name, &starting_list, &service);
    if (script == NULL)
	service = find_service_in_list (name, unavailable_services);
    if (script == NULL)
    {
	int i;
	char txt[1024];

	if ( ( script = calloc (1, sizeof *script) ) == NULL )
	{
	    if (needer != NULL) free (needer);
	    return SIG_FAILED;
	}
	if (service == NULL)
	{
	    service = calloc (1, strlen (name) + sizeof *service);
	    if (service == NULL)
	    {
		free (script);
		if (needer != NULL) free (needer);
		return SIG_FAILED;
	    }
	    strcpy (service->name, name);
	}
	else  /*  Unhook service from unavailable list  */
	{
	    if (service->prev == NULL) unavailable_services = service->next;
	    else service->prev->next = service->next;
	    if (service->next != NULL) service->next->prev = service->prev;
	    service->prev = NULL;
	    service->next = NULL;
	}

	sigprocmask(SIG_BLOCK,&all_signals,&oldsigmask);
	switch ( script->pid = fork () )
	{
	  case 0:   /*  Child   */
	    for (i = 1; i < NSIG; i++) signal (i, SIG_DFL);
	    sigprocmask(SIG_UNBLOCK,&all_signals,NULL);
	    if (!userspace) {
	    	open ("/dev/console", O_RDONLY, 0);
	    	open ("/dev/console", O_RDWR, 0);
	    	dup2 (1, 2);
	    }
	    {
	      const char* fpath=get_path(file);
	      const char* start="start";
	      struct stat Stat;
	      
	      /* if service is a directory, then execute boot prog instead */
	      if ((stat(fpath,&Stat)==0) && (S_ISDIR(Stat.st_mode)))
	      {
	        ssize_t ofs=strcspn(boot_prog," \t\r\n");
	        if (ofs>0) boot_prog[ofs]=0;
	        ystrcat(boot_prog," ");
	        ystrcat(boot_prog,fpath);
	        fpath=boot_prog;
	        start="start-dir";
	      }
	      
	      exec_with_params (fpath, start, NULL);
	      sprintf (txt, "errno %u running program: \"%s\"\n", errno, fpath);
	      err ( _(txt) );
	      _exit (SIG_FAILED);
	    }  
	    break;
	  case -1:  /*  Error   */
	    sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
	    service->next = unavailable_services;
	    service->prev = NULL;
	    if (unavailable_services != NULL)
		unavailable_services->prev = service;
	    unavailable_services = service;
	    free (script);
	    if (needer != NULL) free (needer);
	    return SIG_FAILED;
	    /*break;*/
	  default:  /*  Parent  */
	    sigprocmask(SIG_SETMASK,&oldsigmask,NULL);
	    script->first_service = service;
	    script->last_service = service;
	    insert_entry (&starting_list, script);
	    sched_yield ();
	    break;
	}
    }
    if (needer != NULL) 
    {
    	needer->next = service->needers;
    	service->needers = needer;
    }
    if (waiting_script != NULL) 
    {
      waiting_script->waiting_for = service;
      waiting_script->blocker_pid = pid;
    }  
    return 0;
}   /*  End Function run_command  */

static struct service_struct *find_service_in_list (const char *name,
						    struct service_struct *sv)
{
    for (; sv != NULL; sv = sv->next)
	if (strcmp (sv->name, name) == 0) return (sv);
    return NULL;
}   /*  End Function find_service_in_list  */

/*scans the list referenced by head, every script in this list is checked
if it provides a service name. If so, a ptr to the script is returned and
(unless it is NULL) *service will be set to point to the found service.
If none of the scripts provide name, NULL is returned*/
static struct script_struct *find_script_byname (const char *name,
						 struct list_head *head,
						 struct service_struct **service)
{
    struct script_struct *script;

    for (script = head->first; script != NULL; script = script->next)
    {
	struct service_struct *sv;

	if ( ( sv = find_service_in_list (name, script->first_service) )
	     != NULL )
	{
	    if (service != NULL) *service = sv;
	    return (script);
	}
    }
    if (service != NULL) *service = NULL;
    return NULL;
}   /*  End Function find_script_byname  */

static struct script_struct *find_script_bypid (pid_t pid,
						struct list_head *head)
{
    struct script_struct *script;

    for (script = head->first; script != NULL; script = script->next)
	if (script->pid == pid) return (script);
    return NULL;
}   /*  End Function find_script_bypid  */

/*scans the list referenced by head, every script in this list is checked
if it provides service. If so, a ptr to the script is returned.
If none of the scripts provide service, NULL is returned*/
static struct script_struct *find_script_byservice
		(const struct service_struct* service, struct list_head *head)
{
    struct script_struct *script;

    if (service != NULL)
      for (script = head->first; script != NULL; script = script->next)
      {
	struct service_struct *sv;

	for (sv = script->first_service; sv != NULL; sv = sv->next)
		if (sv == service) return script;
      }
    
    return NULL;
}   /*  End Function find_script_byservice */

/* inserts entry at the beginning of the list referenced by head */
static void insert_entry (struct list_head *head, struct script_struct *entry)
{
    if (entry == NULL) return;
    entry->prev = NULL;
    entry->next = head->first;
    if (head->first != NULL) head->first->prev = entry;
    head->first = entry;
    if (head->last == NULL) head->last = entry;
    ++head->num_entries;
}   /*  End Function insert_entry  */

static void remove_entry (struct list_head *head, struct script_struct *entry)
{
    if (entry->prev == NULL) head->first = entry->next;
    else entry->prev->next = entry->next;
    if (entry->next == NULL) head->last = entry->prev;
    else entry->next->prev = entry->prev;
    --head->num_entries;
}   /*  End Function remove_entry  */

static void signal_needers (struct service_struct *service, int sig)
{
    struct needer_struct *needer, *next_needer;
    for (needer = service->needers; needer != NULL; needer = next_needer)
    {
	kill (needer->pid, sig);
	next_needer = needer->next;
	free (needer);
    }
    service->needers = NULL;
}   /*  End Function signal_needers  */

static void handle_nonworking (struct script_struct *script)
{
    struct service_struct *service, *next;

    for (service = script->first_service; service != NULL; service = next)
    {
	struct script_struct *provider = service->attempting_providers;

	next = service->next;
	if (provider == NULL)
	{
	    service->prev = NULL;
	    service->next = unavailable_services;
	    /* script is free() later so no proper removing from list */
	    if (unavailable_services != NULL)
		unavailable_services->prev = service;
	    unavailable_services = service;
	    
	    /* waiting for unavailable services is not allowed */
	    signal_needers (service, 
			service->failed ? SIG_FAILED : SIG_NOT_PRESENT);
	    clean_up_waiting_for(service);
	    continue;
	}
	service->attempting_providers = provider->next_attempting_provider;
	provider->next_attempting_provider=NULL;
	provider->last_service->next = service;
	service->prev = provider->last_service;
	provider->last_service = service;
	service->next = NULL;
	provider->waiting_for = NULL;
	kill (provider->blocker_pid, SIG_NOT_PRESENT);
    }
    free (script);
}   /*  End Function handle_nonworking  */

static void show_scripts (FILE *fp, const struct script_struct *script,
			  const char *type)
{
    fprintf (fp, "%s SERVICES:\n", type);
    for (; script != NULL; script = script->next)
    {
	struct service_struct *service = script->first_service;

	fputs (service->name, fp);
	for (service = service->next; service != NULL; service = service->next)
	    fprintf (fp, "  (%s)", service->name);
	putc ('\n', fp);
    }
}   /*  End Function show_scripts  */

static const char *get_path (const char *file)
{
#ifdef DEBUG_GETPATH    
    char buffy[1024];
#endif
    char *p1, *p2;
    static char path[PATH_SIZE];

#ifdef DEBUG_GETPATH
    sprintf(buffy,"looking for %s\n",file);
    err(buffy);
#endif
    if (file[0] == '/') return file;
    if (init_path[0] == '\0') return file;
    for (p1 = init_path; *p1 != '\0'; p1 = p2)
    {
	char* p=path;
	int maxlen=PATH_SIZE;
	int len;
	
	if ( ( p2 = strchr (p1, ':') ) == NULL )
	    p2 = p1 + strlen (p1);
	    
	if (*p1 != '/') 
	{
	  len=strlen(script_prefix);
	  if (len>=maxlen) continue;
	  memcpy(p, script_prefix, len);
	  p+=len;
	  maxlen-=len;
	}
	
	len=p2-p1+1; /* +1 for '/' */
	if (len>=maxlen) continue;
	memcpy (p, p1, len-1);
	p[len-1]='/';
	p+=len;
	maxlen-=len;
	
	len=strcspn(file," \t\r\n"); 
	if (len+1>=maxlen) continue; /* +1 because of 0 termination */
        memcpy (p, file, len);
        p[len]=0;
        
	if (*p2 == ':') ++p2;
#ifdef DEBUG_GETPATH
	sprintf(buffy,"checking %s\n",path);
        err(buffy);
#endif
        if (access (path, X_OK) == 0) 
        {
#ifdef DEBUG_GETPATH
           sprintf(buffy,"--> OK -> executing\n");
           err(buffy);
#endif
           return path;
        }
#ifdef DEBUG_GETPATH      
        sprintf(buffy," --> not executable\n");
        err(buffy);
#endif
    }
    return file;
}   /*  End Function get_path  */

void fix_userspace_paths(void)
{
  int len;
  char* p;
  const char* homedir;
#ifdef RELATIVE_TO_HOME
  struct passwd* pw;
  {
    homedir=getenv("HOME");
    if (homedir==NULL)
    {
      pw=getpwuid(getuid());
      if (pw==NULL){
        err ( _("cannot determine home directory\n") );
        exit(1);
      }
      homedir=pw->pw_dir;
    }
  }  
#else
  char buffy[1024];
  {
    homedir=getenv("INIT_ROOT");
    if (homedir==NULL)
    {
      if (getcwd(buffy,1024)==NULL)
      {
        err ( _("cannot determine current directory's name\n") );
        exit(1);
      }
      homedir=buffy;
    }  
  }  
#endif
  len=strlen(homedir);

  setenv("INIT_ROOT",homedir,1);

  ystrcpy(init_root,homedir);
  for (p=strchr(_path_utmp,0); (p+1)!=_path_utmp; --p) p[len]=p[0];
  memcpy(_path_utmp,homedir,len);
  for (p=strchr(_path_wtmp,0); (p+1)!=_path_wtmp; --p) p[len]=p[0];
  memcpy(_path_wtmp,homedir,len);
  for (p=strchr(initctl_name,0); (p+1)!=initctl_name; --p) p[len]=p[0];
  memcpy(initctl_name,homedir,len);
  for (p=strchr(pidfile_name,0); (p+1)!=pidfile_name; --p) p[len]=p[0];
  memcpy(pidfile_name,homedir,len);
  for (p=strchr(_path_inittab,0); (p+1)!=_path_inittab; --p) p[len]=p[0];
  memcpy(_path_inittab,homedir,len);
  for (p=strchr(boot_prog,0); (p+1)!=boot_prog; --p) p[len]=p[0];
  memcpy(boot_prog,homedir,len);
}

/*takes a string like
/bin/sh -c "echo Goodbye World"
and calls execv("/bin/sh",{"sh","-c","echo Hallo World"})
The backslash \ is the escape character inside "...", whenever a 
\ is encountered it will be discarded and the following
char will be left alone. " inside "..." can be written \" and \ inside
"..." can be written \\. The backslash may *not* be used to embed NUL
characters. A backslash preceding a NUL character will be discarded and
the string terminated.
The function will not fail if the closing " are missing. The string is
properly terminated at the 0 character.

argv0 if non-NULL is a string to use for argv[0] (i.e. the program's name)

preargs is a string of arguments to be inserted before the arguments
found in path. May be NULL.

postargs is a string of arguments to be appended after the arguments found
in path. May be NULL.

Example:
exec_with_params("/bin/sh -c \"echo Goodbye World\"","1 2","3 4") will call
execv("/bin/sh",{"sh","1","2","-c","echo Hallo World","3","4"})

RETURN VALUE:
that of execv()
ATTENTION!! The first space will determine the end of the function name
	    and the beginning of the argument list.
*/

int exec_with_params(const char* path,const char* preargs,const char* postargs)
{ 			
  char* argv[MAXPARAMS+1]; /* +1 for final NULL*/
  int argc=0;
  char* st2;
  char* st;
  char buffy[1024]="\0";
  char willow[1024]="\0";
  char oz[1024]="\0";
  char *willowp=willow,*ozp=oz;
  
  ystrcpy(buffy,path);
  if (preargs!=NULL) ystrcpy(willow,preargs);
  if (postargs!=NULL) ystrcpy(oz,postargs);
  
  st=buffy;
  
  if (split_param(&st,&argc,argv,MAXPARAMS)==0) return -1;

  path=argv[0];
  
  /* argv[0] should be just the program name, not a complete path */
  st2=strrchr(argv[0],'/');
  if (st2!=NULL) argv[0]=st2+1;
  
  while(split_param(&willowp,&argc,argv,MAXPARAMS)==1);
  while(split_param(&st,&argc,argv,MAXPARAMS)==1);
  while(split_param(&ozp,&argc,argv,MAXPARAMS)==1);
  
  argv[argc]=NULL;
  return execv(path,argv); 
}

/* This function is useful for splitting a string like
/bin/sh -c "echo Goodbye World"
into {"/bin/sh","-c","echo Hallo World"})

Str is a pointer to a pointer into the string to be split (*Str may be NULL).
The function will skip any leading spaces and set argv[argc] to point to the
first non-space character. Then it will search for the next space character
and replace it with 0. Strings within quotes are handled as a single item.
The backslash is the escape character.
NUL characters can not be embedded in a string, even within quotes.
A backslash preceding a NUL character will be discarded and
the string terminated.
The function will not fail if the closing " are missing. The string is
properly terminated at the 0 character.

*argc is increased after argv[argc] has been written. 
*Str is updated to point after the string just parsed

If *argc>=argc_max, the function does nothing.

RETURN VALUE:
  0 if nothing was parsed (*Str==NULL or *Str is empty except for
  whitespace or *argc>=argc_max)

  1 if a parameter was parsed

ATTENTION!! This function destroys the contents *Str!
*/

int split_param(char** str,int* argc,char* argv[],int argc_max)
{ 			
  char* st;
  int move;
  int insidequotes;
  
  if (str==NULL || *str==NULL || *argc>=argc_max) return 0;
 
  st=*str;
  while (isspace(*st)) ++st;
  
  if (*st==0) return 0;
  
  argv[*argc]=st;
  ++(*argc);
  
  move=0;
  insidequotes=0;
  while(*st!=0 && (!isspace(*st) || insidequotes))
  {
    switch(*st)
    {
      case '"': ++move; ++st; insidequotes=!insidequotes; break;
      case '\\': ++move; ++st; if (*st!=0) {*(st-move)=*st; ++st;} ; break;
      default: *(st-move)=*st; ++st; break;
    }
  }

  if (*st==0) *str=st; else *str=st+1;
  *(st-move)=0;
  return 1;
}

static void clean_up_waiting_for(const struct service_struct* service)
{
  struct script_struct* waits;
  
  if (service == NULL) return;
  
  for (waits = starting_list.first; waits != NULL; waits=waits->next)
  	if (waits->waiting_for == service) waits->waiting_for = NULL;
}

static int waits_for(	const struct script_struct* who, 
			const struct script_struct* forwhat)
{
  if (forwhat != NULL)
  {
    while (who != NULL)
    {
      struct script_struct* waiting_for;
      waiting_for = find_script_byservice(who->waiting_for, &starting_list);
      if (waiting_for == forwhat) return 1;
      who = waiting_for;
    }
  }
  return 0;  
}

static void output_deadlock_scripts(	const struct script_struct* start, 
					const struct script_struct* stop,
					const struct service_struct* service)
{
  const int bufsize=1024;
  char buffy[bufsize];
  
  if ((start != NULL) && (stop != NULL))
  {
    snprintf(buffy,bufsize,"DEADLOCK! %s -> %s\n",
    	stop->first_service->name, service->name);
    err(buffy);
    
    while(start != stop)
    {
      snprintf(buffy,bufsize,"%s -> %s\n",start->first_service->name,
      					  start->waiting_for->name);
      err(buffy);
      start = find_script_byservice(start->waiting_for, &starting_list);
    }
  }  
}
